import json
import os
import traceback
from pathlib import Path

import luigi
from betterboto import client as betterboto_client

from servicecatalog_puppet import constants
import psutil
import logging
import math

logger = logging.getLogger("tasks")
logger.setLevel(logging.INFO)


def unwrap(what):
    if hasattr(what, "get_wrapped"):
        thing = what.get_wrapped()
    else:
        thing = what
    if isinstance(what, dict):
        for k, v in thing.items():
            thing[k] = unwrap(v)
    return thing


class PuppetTask(luigi.Task):
    def read_from_input(self, input_name):
        with self.input().get(input_name).open("r") as f:
            return f.read()

    def load_from_input(self, input_name):
        return json.loads(self.read_from_input(input_name))

    def info(self, message):
        logger.info(f"{self.uid}: {message}")

    def error(self, message):
        logger.error(f"{self.uid}: {message}")

    def warning(self, message):
        logger.warning(f"{self.uid}: {message}")

    def api_calls_used(self):
        return []

    @property
    def resources(self):
        result = {}
        for a in self.api_calls_used():
            result[a] = 1
        return result

    @property
    def output_location(self):
        return f"output/{self.uid}.{self.output_suffix}"

    def remove_output(self):
        os.remove(self.output_location)

    def output(self):
        return luigi.LocalTarget(self.output_location)

    @property
    def output_suffix(self):
        return "json"

    @property
    def uid(self):
        return f"{self.__class__.__name__}/{self.node_id}"

    def params_for_results_display(self):
        return {}

    def write_output(self, content):
        with self.output().open("w") as f:
            f.write(json.dumps(content, indent=4, default=str,))

    @property
    def node_id(self):
        values = [str(v) for v in self.params_for_results_display().values()]
        return "/".join(values)

    def graph_node(self):
        task_friendly_name = self.__class__.__name__.replace("Task", "")
        task_description = ""
        for param, value in self.params_for_results_display().items():
            task_description += f"<br/>{param}: {value}"
        label = f"<b>{task_friendly_name}</b>{task_description}"
        return f'"{self.node_id}" [fillcolor=lawngreen style=filled label= < {label} >]'

    def get_lines(self, haystack):
        lines = []
        if isinstance(haystack, list):
            for i in haystack:
                lines += self.get_lines(i)
        elif isinstance(haystack, dict):
            for i in haystack.values():
                lines += self.get_lines(i)
        else:
            lines.append(f'"{self.node_id}" -> "{haystack.node_id}"')
        return lines

    def get_graph_lines(self):
        return self.get_lines(self.requires())


class GetSSMParamTask(PuppetTask):
    parameter_name = luigi.Parameter()
    name = luigi.Parameter()
    region = luigi.Parameter(default=None)

    cache_invalidator = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "parameter_name": self.parameter_name,
            "name": self.name,
            "region": self.region,
            "cache_invalidator": self.cache_invalidator,
        }

    def api_calls_used(self):
        return ["ssm.get_parameter"]

    def run(self):
        with betterboto_client.ClientContextManager(
            "ssm", region_name=self.region
        ) as ssm:
            try:
                p = ssm.get_parameter(Name=self.name,)
                self.write_output(
                    {
                        "Name": self.name,
                        "Region": self.region,
                        "Value": p.get("Parameter").get("Value"),
                        "Version": p.get("Parameter").get("Version"),
                    }
                )
            except ssm.exceptions.ParameterNotFound as e:
                raise e


def record_event(event_type, task, extra_event_data=None):
    task_type = task.__class__.__name__
    task_params = task.param_kwargs

    event = {
        "event_type": event_type,
        "task_type": task_type,
        "task_params": task_params,
        "params_for_results": task.params_for_results_display(),
    }
    if extra_event_data is not None:
        event.update(extra_event_data)

    with open(
        Path(constants.RESULTS_DIRECTORY)
        / event_type
        / f"{task_type}-{task.task_id}.json",
        "w",
    ) as f:
        f.write(json.dumps(event, default=str, indent=4,))


@luigi.Task.event_handler(luigi.Event.FAILURE)
def on_task_failure(task, exception):
    exception_details = {
        "exception_type": type(exception),
        "exception_stack_trace": traceback.format_exception(
            etype=type(exception), value=exception, tb=exception.__traceback__,
        ),
    }
    record_event("failure", task, exception_details)


def print_stats():
    mem = psutil.virtual_memory()
    logger.info(
        f"memory usage: total={math.ceil(mem.total / 1024 / 1024)}MB used={math.ceil(mem.used / 1024 / 1024)}MB percent={mem.percent}%"
    )


@luigi.Task.event_handler(luigi.Event.SUCCESS)
def on_task_success(task):
    print_stats()
    record_event("success", task)


@luigi.Task.event_handler(luigi.Event.TIMEOUT)
def on_task_timeout(task):
    print_stats()
    record_event("timeout", task)


@luigi.Task.event_handler(luigi.Event.PROCESS_FAILURE)
def on_task_process_failure(task, error_msg):
    print_stats()
    exception_details = {
        "exception_type": "PROCESS_FAILURE",
        "exception_stack_trace": error_msg,
    }
    record_event("process_failure", task, exception_details)


@luigi.Task.event_handler(luigi.Event.PROCESSING_TIME)
def on_task_processing_time(task, duration):
    print_stats()
    record_event("processing_time", task, {"duration": duration})


@luigi.Task.event_handler(luigi.Event.BROKEN_TASK)
def on_task_broken_task(task, exception):
    print_stats()
    record_event("broken_task", task, {"exception": exception})



from . import tasks_unit_tests
from unittest import mock


class DeleteCloudFormationStackTaskTest(tasks_unit_tests.PuppetTaskUnitTest):
    account_id = "0123456789010"
    region = "eu-west-0"
    stack_name = "foo"

    def setUp(self) -> None:
        super().setUp()
        from servicecatalog_puppet.workflow import general

        self.module = general
        self.sut = self.module.DeleteCloudFormationStackTask(
            account_id=self.account_id, region=self.region, stack_name=self.stack_name,
        )

    @mock.patch("servicecatalog_puppet.workflow.general.betterboto_client")
    def test_run(self, mock):
        # setup
        expected_output = "hello world"

        self.expect_call(
            mock,
            dict(
                service="cloudformation",
                role=f"arn:aws:iam::{self.account_id}:role/servicecatalog-puppet/PuppetRole",
                session=f"{self.account_id}-{self.region}-PuppetRole",
                region_name=self.region,
            ),
            "ensure_deleted",
            {"StackName": self.stack_name},
            expected_output,
        )

        # exercise
        self.sut.run()

        # verify
        self.verify(mock)
        self.verify_output(self.sut.params_for_results_display())



import luigi

from servicecatalog_puppet import sdk
from servicecatalog_puppet.workflow import tasks


class BootstrapSpokeAsTask(tasks.PuppetTask):
    puppet_account_id = luigi.Parameter()
    account_id = luigi.Parameter()
    iam_role_arns = luigi.ListParameter()
    role_name = luigi.Parameter()
    permission_boundary = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "account_id": self.account_id,
        }

    def run(self):
        iam_role_arns_to_use = [iam_role_arn for iam_role_arn in self.iam_role_arns]
        iam_role_arns_to_use.append(
            f"arn:aws:iam::{self.account_id}:role/{self.role_name}"
        )
        sdk.bootstrap_spoke_as(
            self.puppet_account_id, iam_role_arns_to_use, self.permission_boundary,
        )
        self.write_output(self.params_for_results_display())



from . import tasks_unit_tests

import pytest


class GeneratePoliciesTemplateTest(tasks_unit_tests.PuppetTaskUnitTest):
    region = "eu-west-0"
    puppet_account_id = "01234567890"
    manifest_file_path = "tcvyuiho"
    cache_invalidator = "foo"

    def setUp(self) -> None:
        from . import generate

        self.sut = generate.GeneratePoliciesTemplate(
            puppet_account_id=self.puppet_account_id,
            manifest_file_path=self.manifest_file_path,
            region=self.region,
            sharing_policies=dict(),
            cache_invalidator=self.cache_invalidator,
        )

    def test_params_for_results_display(self):
        expected_result = {
            "manifest_file_path": self.manifest_file_path,
            "puppet_account_id": self.puppet_account_id,
            "region": self.region,
            "cache_invalidator": self.cache_invalidator,
        }
        self.assertEqual(expected_result, self.sut.params_for_results_display())


class EnsureEventBridgeEventBusTaskTest(tasks_unit_tests.PuppetTaskUnitTest):
    region = "eu-west-0"
    puppet_account_id = "01234567890"

    def setUp(self) -> None:
        from . import generate

        self.sut = generate.EnsureEventBridgeEventBusTask(
            region=self.region, puppet_account_id=self.puppet_account_id,
        )

    def test_params_for_results_display(self):
        expected_result = {
            "region": self.region,
            "puppet_account_id": self.puppet_account_id,
        }
        self.assertEqual(expected_result, self.sut.params_for_results_display())


class GeneratePoliciesTest(tasks_unit_tests.PuppetTaskUnitTest):
    region = "eu-west-0"
    puppet_account_id = "01234567890"
    manifest_file_path = "tcvyuiho"
    sharing_policies = dict()

    should_use_sns = False
    should_use_product_plans = True
    include_expanded_from = True
    single_account = None
    is_dry_run = False
    execution_mode = "hub"
    cache_invalidator = "foo"

    def setUp(self) -> None:
        from . import generate

        self.sut = generate.GeneratePolicies(
            puppet_account_id=self.puppet_account_id,
            manifest_file_path=self.manifest_file_path,
            region=self.region,
            sharing_policies=self.sharing_policies,
            should_use_sns=self.should_use_sns,
            cache_invalidator=self.cache_invalidator,
        )

    def test_params_for_results_display(self):
        expected_result = {
            "manifest_file_path": self.manifest_file_path,
            "puppet_account_id": self.puppet_account_id,
            "region": self.region,
            "should_use_sns": self.should_use_sns,
            "cache_invalidator": self.cache_invalidator,
        }
        self.assertEqual(expected_result, self.sut.params_for_results_display())

    def test_requires(self):
        from . import generate

        expected_result = {
            "template": generate.GeneratePoliciesTemplate(
                puppet_account_id=self.puppet_account_id,
                manifest_file_path=self.manifest_file_path,
                region=self.region,
                sharing_policies=dict(),
                cache_invalidator="foo",
            ),
        }
        self.assertEqual(expected_result, self.sut.requires())


class GenerateSharesTaskTest(tasks_unit_tests.PuppetTaskUnitTest):
    puppet_account_id = "01234567890"
    manifest_file_path = "tcvyuiho"

    should_use_sns = False
    should_use_product_plans = True
    include_expanded_from = True
    single_account = None
    is_dry_run = False
    execution_mode = "hub"

    section = "launches"
    cache_invalidator = "food"

    def setUp(self) -> None:
        from . import generate

        self.sut = generate.GenerateSharesTask(
            puppet_account_id=self.puppet_account_id,
            manifest_file_path=self.manifest_file_path,
            should_use_sns=self.should_use_sns,
            section=self.section,
            cache_invalidator=self.cache_invalidator,
        )

    def test_params_for_results_display(self):
        expected_result = {
            "puppet_account_id": self.puppet_account_id,
            "manifest_file_path": self.manifest_file_path,
            "section": self.section,
            "cache_invalidator": self.cache_invalidator,
        }
        self.assertEqual(expected_result, self.sut.params_for_results_display())



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import logging

logger = logging.getLogger(__name__)


def ext_schema(value, rule_obj, path):
    ensure_deploy_to_tags_exist(value)
    ensure_depends_on_exist(value)
    return True


def ensure_deploy_to_tags_exist(value):
    tags_defined_by_accounts = {}
    for account in value.get("accounts"):
        for tag in account.get("tags"):
            tags_defined_by_accounts[tag] = True

    for collection_type in ["launches", "spoke-local-portfolios"]:
        collection_to_check = value.get(collection_type, [])
        for collection_name, collection_item in collection_to_check.items():
            for deploy_to in collection_item.get("deploy_to", {}).get("tags", []):
                tag_to_check = deploy_to.get("tag")
                if tags_defined_by_accounts.get(tag_to_check) is None:
                    raise AssertionError(
                        f"{collection_type}.{collection_name} uses tag {tag_to_check} in deploy_to that does not exist"
                    )


def ensure_depends_on_exist(value):
    for collection_type in ["launches", "spoke-local-portfolios"]:
        collection_to_check = value.get(collection_type, [])
        for collection_name, collection_item in collection_to_check.items():
            for depends_on in collection_item.get("depends_on", []):
                if value.get("launches").get(depends_on) is None:
                    raise AssertionError(
                        f"{collection_type}.{collection_name} uses {depends_on} in depends_on that does not exist"
                    )



import json
from functools import lru_cache

import luigi

from servicecatalog_puppet import config, constants

from servicecatalog_puppet.workflow import manifest as manifest_tasks
from servicecatalog_puppet.workflow import (
    portfoliomanagement as portfoliomanagement_tasks,
)
from servicecatalog_puppet.workflow import general as general_tasks
from servicecatalog_puppet.workflow import tasks
from betterboto import client as betterboto_client


class GeneratePoliciesTemplate(tasks.PuppetTask):
    puppet_account_id = luigi.Parameter()
    manifest_file_path = luigi.Parameter()
    region = luigi.Parameter()
    sharing_policies = luigi.DictParameter()

    cache_invalidator = luigi.Parameter()

    @property
    def output_suffix(self):
        return "template.yaml"

    def params_for_results_display(self):
        return {
            "manifest_file_path": self.manifest_file_path,
            "puppet_account_id": self.puppet_account_id,
            "region": self.region,
            "cache_invalidator": self.cache_invalidator,
        }

    def run(self):
        rendered = config.env.get_template("policies.template.yaml.j2").render(
            sharing_policies=self.sharing_policies,
            VERSION=config.get_puppet_version(),
            HOME_REGION=self.region,
        )
        with self.output().open("w") as output_file:
            output_file.write(rendered)


class EnsureEventBridgeEventBusTask(tasks.PuppetTask):
    puppet_account_id = luigi.Parameter()
    region = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "region": self.region,
        }

    def api_calls_used(self):
        return {
            f"events.describe_event_bus_{self.puppet_account_id}_{self.region}": 1,
            f"events.create_event_bus_{self.puppet_account_id}_{self.region}": 1,
        }

    def run(self):
        with betterboto_client.CrossAccountClientContextManager(
            "events",
            f"arn:aws:iam::{self.puppet_account_id}:role/servicecatalog-puppet/PuppetRole",
            f"events-{self.puppet_account_id}-{self.region}",
            region_name=self.region,
        ) as events:
            created = False
            try:
                events.describe_event_bus(Name=constants.EVENT_BUS_NAME)
                created = True
            except events.exceptions.ResourceNotFoundException:
                events.create_event_bus(Name=constants.EVENT_BUS_NAME)
        self.write_output(created)


class GeneratePolicies(tasks.PuppetTask):
    puppet_account_id = luigi.Parameter()
    manifest_file_path = luigi.Parameter()
    region = luigi.Parameter()
    sharing_policies = luigi.DictParameter()
    should_use_sns = luigi.BoolParameter()

    cache_invalidator = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "manifest_file_path": self.manifest_file_path,
            "puppet_account_id": self.puppet_account_id,
            "region": self.region,
            "should_use_sns": self.should_use_sns,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        return {
            "template": GeneratePoliciesTemplate(
                puppet_account_id=self.puppet_account_id,
                manifest_file_path=self.manifest_file_path,
                region=self.region,
                sharing_policies=self.sharing_policies,
                cache_invalidator=self.cache_invalidator,
            ),
        }

    def api_calls_used(self):
        return {
            f"cloudformation.create_or_update_{self.puppet_account_id}_{self.region}": 1,
        }

    def run(self):
        template = self.read_from_input("template")
        with betterboto_client.CrossAccountClientContextManager(
            "cloudformation",
            f"arn:aws:iam::{self.puppet_account_id}:role/servicecatalog-puppet/PuppetRole",
            f"cf-{self.puppet_account_id}-{self.region}",
            region_name=self.region,
        ) as cloudformation:
            cloudformation.create_or_update(
                ShouldUseChangeSets=False,
                StackName="servicecatalog-puppet-policies",
                TemplateBody=template,
                NotificationARNs=[
                    f"arn:aws:sns:{self.region}:{self.puppet_account_id}:servicecatalog-puppet-cloudformation-regional-events"
                ]
                if self.should_use_sns
                else [],
            )
        self.write_output(self.get_sharing_policies())

    @lru_cache()
    def get_sharing_policies(self):
        return json.loads(json.dumps(self.sharing_policies.get_wrapped()))


class GenerateSharesTask(tasks.PuppetTask, manifest_tasks.ManifestMixen):
    puppet_account_id = luigi.Parameter()
    manifest_file_path = luigi.Parameter()
    should_use_sns = luigi.BoolParameter()
    section = luigi.Parameter()
    cache_invalidator = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "manifest_file_path": self.manifest_file_path,
            "section": self.section,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        portfolios = dict()
        requirements = dict(
            deletes=list(),
            ensure_event_buses=list(),
            generate_policies=list(),
            portfolios=portfolios,
        )
        for region_name, accounts in self.manifest.get_accounts_by_region().items():
            requirements["deletes"].append(
                general_tasks.DeleteCloudFormationStackTask(
                    account_id=self.puppet_account_id,
                    region=region_name,
                    stack_name="servicecatalog-puppet-shares",
                )
            )

        for (
            region_name,
            sharing_policies,
        ) in self.manifest.get_sharing_policies_by_region().items():
            requirements["ensure_event_buses"].append(
                EnsureEventBridgeEventBusTask(
                    puppet_account_id=self.puppet_account_id, region=region_name,
                )
            )

            requirements["generate_policies"].append(
                GeneratePolicies(
                    puppet_account_id=self.puppet_account_id,
                    manifest_file_path=self.manifest_file_path,
                    region=region_name,
                    sharing_policies=sharing_policies,
                    should_use_sns=self.should_use_sns,
                    cache_invalidator=self.cache_invalidator,
                )
            )

        for (
            region_name,
            shares_by_portfolio_account,
        ) in self.manifest.get_shares_by_region_portfolio_account(
            self.puppet_account_id, self.section
        ).items():
            for (
                portfolio_name,
                shares_by_account,
            ) in shares_by_portfolio_account.items():
                for account_id, share in shares_by_account.items():
                    i = "_".join(
                        [
                            str(self.puppet_account_id),
                            portfolio_name,
                            str(account_id),
                            region_name,
                        ]
                    )
                    portfolios[
                        i
                    ] = portfoliomanagement_tasks.GetPortfolioByPortfolioName(
                        manifest_file_path=self.manifest_file_path,
                        puppet_account_id=self.puppet_account_id,
                        portfolio=portfolio_name,
                        account_id=self.puppet_account_id,
                        region=region_name,
                        cache_invalidator=self.cache_invalidator,
                    )
        return requirements

    def run(self):
        tasks = list()
        for (
            region_name,
            shares_by_portfolio_account,
        ) in self.manifest.get_shares_by_region_portfolio_account(
            self.puppet_account_id, self.section
        ).items():
            for (
                portfolio_name,
                shares_by_account,
            ) in shares_by_portfolio_account.items():
                for account_id, share in shares_by_account.items():
                    i = "_".join(
                        [
                            str(self.puppet_account_id),
                            portfolio_name,
                            str(account_id),
                            region_name,
                        ]
                    )
                    portfolio_input = self.input().get("portfolios").get(i)

                    if portfolio_input is None:
                        raise Exception(
                            f"failed to get portfolios details for {i} in {self.input().get('portfolios')}"
                        )

                    portfolio = json.loads(portfolio_input.open("r").read())

                    tasks.append(
                        portfoliomanagement_tasks.CreateShareForAccountLaunchRegion(
                            manifest_file_path=self.manifest_file_path,
                            puppet_account_id=self.puppet_account_id,
                            account_id=account_id,
                            region=region_name,
                            portfolio=portfolio_name,
                            portfolio_id=portfolio.get("portfolio_id"),
                            sharing_mode=share.get(self.section).get(
                                "sharing_mode",
                                config.get_global_sharing_mode_default(
                                    self.puppet_account_id
                                ),
                            ),
                        )
                    )

        yield tasks
        self.write_output(self.params_for_results_display())



# from . import tasks_unit_tests
#
#
# class SpokeLocalPortfolioSectionTaskTest(tasks_unit_tests.PuppetTaskUnitTest):
#
#     def setUp(self) -> None:
#         from . import spoke_local_portfolios
#
#         self.sut = spoke_local_portfolios.SpokeLocalPortfolioSectionTask(
#         )
#
#     def test_params_for_results_display(self):
#         expected_result = {
#             "puppet_account_id": self.puppet_account_id,
#             "manifest_file_path": self.manifest_file_path,
#         }
#         self.assertEqual(expected_result, self.sut.params_for_results_display())



import json
import os
from functools import lru_cache

import luigi
from betterboto import client as betterboto_client
from servicecatalog_puppet.workflow import generate as generate_tasks

from servicecatalog_puppet import manifest_utils_for_launches
from servicecatalog_puppet import aws
from servicecatalog_puppet import config

from servicecatalog_puppet import constants
from servicecatalog_puppet import manifest_utils
from servicecatalog_puppet.workflow import tasks
from servicecatalog_puppet.workflow import (
    portfoliomanagement as portfoliomanagement_tasks,
)
from servicecatalog_puppet.workflow import manifest as manifest_tasks


class ProvisioningTask(tasks.PuppetTask):
    manifest_file_path = luigi.Parameter()


class ProvisioningArtifactParametersTask(ProvisioningTask):
    puppet_account_id = luigi.Parameter()
    portfolio = luigi.Parameter()
    portfolio_id = luigi.Parameter()
    product = luigi.Parameter()
    product_id = luigi.Parameter()
    version = luigi.Parameter()
    version_id = luigi.Parameter()
    account_id = luigi.Parameter()
    region = luigi.Parameter()

    @property
    def retry_count(self):
        return 5

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "portfolio_id": self.portfolio_id,
            "product": self.product,
            "product_id": self.product_id,
            "version": self.version,
            "version_id": self.version_id,
            "account_id": self.account_id,
            "region": self.region,
        }

    def api_calls_used(self):
        return [
            f"servicecatalog.list_launch_paths_{self.account_id}_{self.region}",
            f"servicecatalog.describe_provisioning_parameters_{self.account_id}_{self.region}",
        ]

    def run(self):
        with betterboto_client.CrossAccountClientContextManager(
            "servicecatalog",
            f"arn:aws:iam::{self.account_id}:role/servicecatalog-puppet/PuppetRole",
            f"{self.account_id}-{self.region}-sc",
            region_name=self.region,
        ) as service_catalog:
            self.info(
                f"getting path for {self.product_id} of portfolio: {self.portfolio}"
            )
            path_id = aws.get_path_for_product(
                service_catalog, self.product_id, self.portfolio
            )
            provisioning_artifact_parameters = service_catalog.describe_provisioning_parameters(
                ProductId=self.product_id,
                ProvisioningArtifactId=self.version_id,
                PathId=path_id,
            ).get(
                "ProvisioningArtifactParameters", []
            )
            self.write_output(
                provisioning_artifact_parameters
                if isinstance(provisioning_artifact_parameters, list)
                else [provisioning_artifact_parameters]
            )


class ProvisionProductTask(ProvisioningTask):
    launch_name = luigi.Parameter()
    portfolio = luigi.Parameter()
    portfolio_id = luigi.Parameter()
    product = luigi.Parameter()
    product_id = luigi.Parameter()
    version = luigi.Parameter()
    version_id = luigi.Parameter()
    region = luigi.Parameter()
    account_id = luigi.Parameter()

    puppet_account_id = luigi.Parameter()

    parameters = luigi.ListParameter(default=[], significant=False)
    ssm_param_inputs = luigi.ListParameter(default=[], significant=False)

    launch_parameters = luigi.DictParameter(default={}, significant=False)
    manifest_parameters = luigi.DictParameter(default={}, significant=False)
    account_parameters = luigi.DictParameter(default={}, significant=False)

    retry_count = luigi.IntParameter(default=1, significant=False)
    worker_timeout = luigi.IntParameter(default=0, significant=False)
    ssm_param_outputs = luigi.ListParameter(default=[], significant=False)
    should_use_sns = luigi.BoolParameter(significant=False, default=False)
    should_use_product_plans = luigi.BoolParameter(significant=False, default=False)
    requested_priority = luigi.IntParameter(significant=False, default=0)

    execution = luigi.Parameter()

    cache_invalidator = luigi.Parameter()

    try_count = 1
    all_params = []

    def params_for_results_display(self):
        return {
            "launch_name": self.launch_name,
            "account_id": self.account_id,
            "region": self.region,
            "portfolio": self.portfolio,
            "portfolio_id": self.portfolio_id,
            "product": self.product,
            "product_id": self.product_id,
            "version": self.version,
            "version_id": self.version_id,
            "execution": self.execution,
            "cache_invalidator": self.cache_invalidator,
        }

    @property
    def priority(self):
        return self.requested_priority

    def requires(self):
        all_params = {}
        all_params.update(self.manifest_parameters)
        all_params.update(self.launch_parameters)
        all_params.update(self.account_parameters)

        ssm_params = {}

        for param_name, param_details in all_params.items():
            if param_details.get("ssm"):
                if param_details.get("default"):
                    del param_details["default"]
                ssm_params[param_name] = tasks.GetSSMParamTask(
                    parameter_name=param_name,
                    name=param_details.get("ssm").get("name"),
                    region=param_details.get("ssm").get(
                        "region", config.get_home_region(self.puppet_account_id)
                    ),
                    cache_invalidator=self.cache_invalidator,
                )
        self.all_params = all_params

        return {
            "ssm_params": ssm_params,
            "provisioning_artifact_parameters": ProvisioningArtifactParametersTask(
                manifest_file_path=self.manifest_file_path,
                puppet_account_id=self.puppet_account_id,
                portfolio=self.portfolio,
                portfolio_id=self.portfolio_id,
                product=self.product,
                product_id=self.product_id,
                version=self.version,
                version_id=self.version_id,
                account_id=self.account_id,
                region=self.region,
            ),
        }

    def api_calls_used(self):
        return {
            f"servicecatalog.list_launch_paths_{self.account_id}_{self.region}",
            f"servicecatalog.scan_provisioned_products_single_page_{self.account_id}_{self.region}",
            f"servicecatalog.describe_provisioned_product_{self.account_id}_{self.region}",
            f"servicecatalog.terminate_provisioned_product_{self.account_id}_{self.region}",
            f"servicecatalog.describe_record_{self.account_id}_{self.region}",
            f"cloudformation.get_template_summary_{self.account_id}_{self.region}",
            f"cloudformation.describe_stacks_{self.account_id}_{self.region}",
            f"servicecatalog.list_provisioned_product_plans_single_page_{self.account_id}_{self.region}",
            f"servicecatalog.delete_provisioned_product_plan_{self.account_id}_{self.region}",
            f"servicecatalog.create_provisioned_product_plan_{self.account_id}_{self.region}",
            f"servicecatalog.describe_provisioned_product_plan_{self.account_id}_{self.region}",
            f"servicecatalog.execute_provisioned_product_plan_{self.account_id}_{self.region}",
            f"servicecatalog.describe_provisioned_product_{self.account_id}_{self.region}",
            f"servicecatalog.update_provisioned_product_{self.account_id}_{self.region}",
            f"servicecatalog.provision_product_{self.account_id}_{self.region}",
            # f"ssm.put_parameter_and_wait_{self.region}",
        }

    def run(self):
        ssm_params = dict()
        task_output = dict(
            **self.params_for_results_display(),
            account_parameters=tasks.unwrap(self.account_parameters),
            launch_parameters=tasks.unwrap(self.launch_parameters),
            manifest_parameters=tasks.unwrap(self.manifest_parameters),
            ssm_params=ssm_params,
        )
        for k, v in self.input().get("ssm_params").items():
            ssm_params[k] = json.loads(v.open("r").read())

        all_params = self.get_all_params()

        role = f"arn:aws:iam::{self.account_id}:role/servicecatalog-puppet/PuppetRole"
        with betterboto_client.CrossAccountClientContextManager(
            "servicecatalog",
            role,
            f"sc-{self.region}-{self.account_id}",
            region_name=self.region,
        ) as service_catalog:
            path_id = aws.get_path_for_product(
                service_catalog, self.product_id, self.portfolio
            )

            (
                provisioned_product_id,
                provisioning_artifact_id,
            ) = aws.terminate_if_status_is_not_available(
                service_catalog,
                self.launch_name,
                self.product_id,
                self.account_id,
                self.region,
            )
            self.info(
                f"pp_id: {provisioned_product_id}, paid : {provisioning_artifact_id}"
            )

            with betterboto_client.CrossAccountClientContextManager(
                "cloudformation",
                role,
                f"cfn-{self.region}-{self.account_id}",
                region_name=self.region,
            ) as cloudformation:
                need_to_provision = True

                self.info(
                    f"running as {role},checking {self.product_id} {self.version_id} {path_id} in {self.account_id} {self.region}"
                )

                with self.input().get("provisioning_artifact_parameters").open(
                    "r"
                ) as f:
                    provisioning_artifact_parameters = json.loads(f.read())

                params_to_use = {}
                for p in provisioning_artifact_parameters:
                    param_name = p.get("ParameterKey")
                    params_to_use[param_name] = all_params.get(
                        param_name, p.get("DefaultValue")
                    )

                if provisioning_artifact_id == self.version_id:
                    self.info(f"found previous good provision")
                    if provisioned_product_id:
                        self.info(f"checking params for diffs")
                        provisioned_parameters = aws.get_parameters_for_stack(
                            cloudformation,
                            f"SC-{self.account_id}-{provisioned_product_id}",
                        )
                        self.info(f"current params: {provisioned_parameters}")

                        self.info(f"new params: {params_to_use}")

                        if provisioned_parameters == params_to_use:
                            self.info(f"params unchanged")
                            need_to_provision = False
                        else:
                            self.info(f"params changed")

                if need_to_provision:
                    self.info(
                        f"about to provision with params: {json.dumps(params_to_use)}"
                    )

                    if provisioned_product_id:
                        stack = aws.get_stack_output_for(
                            cloudformation,
                            f"SC-{self.account_id}-{provisioned_product_id}",
                        )
                        stack_status = stack.get("StackStatus")
                        self.info(f"current cfn stack_status is {stack_status}")
                        if stack_status not in [
                            "UPDATE_COMPLETE",
                            "CREATE_COMPLETE",
                            "UPDATE_ROLLBACK_COMPLETE",
                        ]:
                            raise Exception(
                                f"[{self.uid}] current cfn stack_status is {stack_status}"
                            )
                        if stack_status == "UPDATE_ROLLBACK_COMPLETE":
                            self.warning(
                                f"[{self.uid}] SC-{self.account_id}-{provisioned_product_id} has a status of "
                                f"{stack_status}.  This may need manual resolution."
                            )

                    if provisioned_product_id:
                        if self.should_use_product_plans:
                            provisioned_product_id = aws.provision_product_with_plan(
                                service_catalog,
                                self.launch_name,
                                self.account_id,
                                self.region,
                                self.product_id,
                                self.version_id,
                                self.puppet_account_id,
                                path_id,
                                params_to_use,
                                self.version,
                                self.should_use_sns,
                            )
                        else:
                            provisioned_product_id = aws.update_provisioned_product(
                                service_catalog,
                                self.launch_name,
                                self.account_id,
                                self.region,
                                self.product_id,
                                self.version_id,
                                self.puppet_account_id,
                                path_id,
                                params_to_use,
                                self.version,
                                self.execution,
                            )

                    else:
                        provisioned_product_id = aws.provision_product(
                            service_catalog,
                            self.launch_name,
                            self.account_id,
                            self.region,
                            self.product_id,
                            self.version_id,
                            self.puppet_account_id,
                            path_id,
                            params_to_use,
                            self.version,
                            self.should_use_sns,
                            self.execution,
                        )

                self.info(f"self.execution_mode is {self.execution}")
                if self.execution == constants.EXECUTION_MODE_HUB:
                    self.info(
                        f"Running in execution mode: {self.execution}, checking for SSM outputs"
                    )
                    with betterboto_client.CrossAccountClientContextManager(
                        "cloudformation",
                        role,
                        f"cfn-{self.region}-{self.account_id}",
                        region_name=self.region,
                    ) as spoke_cloudformation:
                        stack_details = aws.get_stack_output_for(
                            spoke_cloudformation,
                            f"SC-{self.account_id}-{provisioned_product_id}",
                        )

                    for ssm_param_output in self.ssm_param_outputs:
                        self.info(
                            f"writing SSM Param: {ssm_param_output.get('stack_output')}"
                        )
                        with betterboto_client.ClientContextManager("ssm") as ssm:
                            found_match = False
                            # TODO push into another task
                            for output in stack_details.get("Outputs", []):
                                if output.get("OutputKey") == ssm_param_output.get(
                                    "stack_output"
                                ):
                                    found_match = True
                                    self.info(f"found value")
                                    ssm.put_parameter_and_wait(
                                        Name=ssm_param_output.get("param_name"),
                                        Value=output.get("OutputValue"),
                                        Type=ssm_param_output.get(
                                            "param_type", "String"
                                        ),
                                        Overwrite=True,
                                    )
                            if not found_match:
                                raise Exception(
                                    f"[{self.uid}] Could not find match for {ssm_param_output.get('stack_output')}"
                                )

                    self.write_output(task_output)
                else:
                    self.write_output(task_output)
                self.info("finished")

    def get_all_params(self):
        all_params = {}
        self.info(f"collecting all_params")
        for param_name, param_details in self.all_params.items():
            if param_details.get("ssm"):
                with self.input().get("ssm_params").get(param_name).open() as f:
                    all_params[param_name] = json.loads(f.read()).get("Value")
            if param_details.get("default"):
                all_params[param_name] = param_details.get("default")
        self.info(f"finished collecting all_params: {all_params}")
        return all_params


class ProvisionProductDryRunTask(ProvisionProductTask):
    def api_calls_used(self):
        return [
            f"servicecatalog.scan_provisioned_products_single_page_{self.account_id}_{self.region}",
            f"servicecatalog.list_launch_paths_{self.account_id}_{self.region}",
            f"servicecatalog.describe_provisioning_artifact_{self.account_id}_{self.region}",
            f"cloudformation.describe_provisioning_artifact_{self.account_id}_{self.region}",
            f"cloudformation.get_template_summary_{self.account_id}_{self.region}",
            f"cloudformation.describe_stacks_{self.account_id}_{self.region}",
        ]

    def run(self):
        self.info(f"starting deploy try {self.try_count} of {self.retry_count}")

        all_params = self.get_all_params()

        role = f"arn:aws:iam::{self.account_id}:role/servicecatalog-puppet/PuppetRole"
        with betterboto_client.CrossAccountClientContextManager(
            "servicecatalog",
            role,
            f"sc-{self.region}-{self.account_id}",
            region_name=self.region,
        ) as service_catalog:
            self.info(f"looking for previous failures")
            path_id = aws.get_path_for_product(
                service_catalog, self.product_id, self.portfolio
            )

            response = service_catalog.scan_provisioned_products_single_page(
                AccessLevelFilter={"Key": "Account", "Value": "self"},
            )

            provisioned_product_id = False
            provisioning_artifact_id = False
            current_status = "NOT_PROVISIONED"
            for r in response.get("ProvisionedProducts", []):
                if r.get("Name") == self.launch_name:
                    current_status = r.get("Status")
                    if current_status in ["AVAILABLE", "TAINTED"]:
                        provisioned_product_id = r.get("Id")
                        provisioning_artifact_id = r.get("ProvisioningArtifactId")

            self.info(
                f"pp_id: {provisioned_product_id}, paid : {provisioning_artifact_id}"
            )
            current_version_details = self.get_current_version(
                provisioning_artifact_id, service_catalog
            )

            with betterboto_client.CrossAccountClientContextManager(
                "cloudformation",
                role,
                f"cfn-{self.region}-{self.account_id}",
                region_name=self.region,
            ) as cloudformation:
                self.info(
                    f"running as {role},checking {self.product_id} {self.version_id} {path_id} in {self.account_id} {self.region}"
                )

                with self.input().get("provisioning_artifact_parameters").open(
                    "r"
                ) as f:
                    provisioning_artifact_parameters = json.loads(f.read())

                params_to_use = {}
                for p in provisioning_artifact_parameters:
                    param_name = p.get("ParameterKey")
                    params_to_use[param_name] = all_params.get(
                        param_name, p.get("DefaultValue")
                    )

                if provisioning_artifact_id == self.version_id:
                    self.info(f"found previous good provision")
                    if provisioned_product_id:
                        self.info(f"checking params for diffs")
                        provisioned_parameters = aws.get_parameters_for_stack(
                            cloudformation,
                            f"SC-{self.account_id}-{provisioned_product_id}",
                        )
                        self.info(f"current params: {provisioned_parameters}")
                        self.info(f"new params: {params_to_use}")

                        if provisioned_parameters == params_to_use:
                            self.info(f"params unchanged")
                            self.write_result(
                                current_version=self.version,
                                new_version=self.version,
                                effect=constants.NO_CHANGE,
                                current_status=current_status,
                                active=current_version_details.get("Active"),
                                notes="Versions and params are the same",
                            )
                        else:
                            self.write_result(
                                current_version=self.version,
                                new_version=self.version,
                                effect=constants.CHANGE,
                                current_status=current_status,
                                active=current_version_details.get("Active"),
                                notes="Versions are the same but the params are different",
                            )
                else:
                    if provisioning_artifact_id:
                        current_version = current_version_details.get("Name")
                        active = current_version_details.get("Active")
                    else:
                        current_version = ""
                        active = False
                    self.write_result(
                        current_version=current_version,
                        new_version=self.version,
                        effect=constants.CHANGE,
                        current_status=current_status,
                        active=active,
                        notes="Version change",
                    )

    def get_current_version(self, provisioning_artifact_id, service_catalog):
        return service_catalog.describe_provisioning_artifact(
            ProvisioningArtifactId=provisioning_artifact_id, ProductId=self.product_id,
        ).get("ProvisioningArtifactDetail")

    def write_result(
        self, current_version, new_version, effect, current_status, active, notes=""
    ):
        with self.output().open("w") as f:
            f.write(
                json.dumps(
                    {
                        "current_version": current_version,
                        "new_version": new_version,
                        "effect": effect,
                        "current_status": current_status,
                        "active": active,
                        "notes": notes,
                        "params": self.param_kwargs,
                    },
                    indent=4,
                    default=str,
                )
            )


class TerminateProductTask(ProvisioningTask):
    launch_name = luigi.Parameter()
    portfolio = luigi.Parameter()
    portfolio_id = luigi.Parameter()
    product = luigi.Parameter()
    product_id = luigi.Parameter()
    version = luigi.Parameter()
    version_id = luigi.Parameter()

    account_id = luigi.Parameter()
    region = luigi.Parameter()
    puppet_account_id = luigi.Parameter()

    retry_count = luigi.IntParameter(default=1)

    ssm_param_outputs = luigi.ListParameter(default=[])

    worker_timeout = luigi.IntParameter(default=0, significant=False)

    try_count = 1

    parameters = luigi.ListParameter(default=[])
    ssm_param_inputs = luigi.ListParameter(default=[])

    # dependencies = luigi.ListParameter(default=[])

    cache_invalidator = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "launch_name": self.launch_name,
            "account_id": self.account_id,
            "region": self.region,
            "portfolio": self.portfolio,
            "portfolio_id": self.portfolio_id,
            "product": self.product,
            "product_id": self.product_id,
            "version": self.version,
            "version_id": self.version_id,
            "cache_invalidator": self.cache_invalidator,
        }

    def api_calls_used(self):
        return [
            f"servicecatalog.scan_provisioned_products_single_page{self.account_id}_{self.region}",
            f"servicecatalog.terminate_provisioned_product_{self.account_id}_{self.region}",
            f"servicecatalog.describe_record_{self.account_id}_{self.region}",
            # f"ssm.delete_parameter_{self.region}": 1,
        ]

    def run(self):
        self.info(f"starting terminate try {self.try_count} of {self.retry_count}")

        role = f"arn:aws:iam::{self.account_id}:role/servicecatalog-puppet/PuppetRole"
        with betterboto_client.CrossAccountClientContextManager(
            "servicecatalog",
            role,
            f"sc-{self.region}-{self.account_id}",
            region_name=self.region,
        ) as service_catalog:
            self.info(
                f"[{self.launch_name}] {self.account_id}:{self.region} :: looking for previous failures"
            )
            provisioned_product_id, provisioning_artifact_id = aws.ensure_is_terminated(
                service_catalog, self.launch_name, self.product_id
            )
            log_output = self.to_str_params()
            log_output.update(
                {"provisioned_product_id": provisioned_product_id,}
            )

            for ssm_param_output in self.ssm_param_outputs:
                param_name = ssm_param_output.get("param_name")
                self.info(
                    f"[{self.launch_name}] {self.account_id}:{self.region} :: deleting SSM Param: {param_name}"
                )
                with betterboto_client.ClientContextManager("ssm") as ssm:
                    try:
                        # todo push into another task
                        ssm.delete_parameter(Name=param_name,)
                        self.info(
                            f"[{self.launch_name}] {self.account_id}:{self.region} :: deleting SSM Param: {param_name}"
                        )
                    except ssm.exceptions.ParameterNotFound:
                        self.info(
                            f"[{self.launch_name}] {self.account_id}:{self.region} :: SSM Param: {param_name} not found"
                        )

            with self.output().open("w") as f:
                f.write(json.dumps(log_output, indent=4, default=str,))

            self.info(
                f"[{self.launch_name}] {self.account_id}:{self.region} :: finished terminating"
            )


class TerminateProductDryRunTask(ProvisioningTask):
    launch_name = luigi.Parameter()
    portfolio = luigi.Parameter()
    portfolio_id = luigi.Parameter()
    product = luigi.Parameter()
    product_id = luigi.Parameter()
    version = luigi.Parameter()
    version_id = luigi.Parameter()

    account_id = luigi.Parameter()
    region = luigi.Parameter()
    puppet_account_id = luigi.Parameter()

    retry_count = luigi.IntParameter(default=1)

    ssm_param_outputs = luigi.ListParameter(default=[])

    worker_timeout = luigi.IntParameter(default=0, significant=False)

    parameters = luigi.ListParameter(default=[])
    ssm_param_inputs = luigi.ListParameter(default=[])

    cache_invalidator = luigi.Parameter()

    try_count = 1

    def params_for_results_display(self):
        return {
            "launch_name": self.launch_name,
            "account_id": self.account_id,
            "region": self.region,
            "portfolio": self.portfolio,
            "portfolio_id": self.portfolio_id,
            "product": self.product,
            "product_id": self.product_id,
            "version": self.version,
            "version_id": self.version_id,
            "cache_invalidator": self.cache_invalidator,
        }

    def write_result(self, current_version, new_version, effect, notes=""):
        with self.output().open("w") as f:
            f.write(
                json.dumps(
                    {
                        "current_version": current_version,
                        "new_version": new_version,
                        "effect": effect,
                        "notes": notes,
                        "params": self.param_kwargs,
                    },
                    indent=4,
                    default=str,
                )
            )

    def api_calls_used(self):
        return [
            f"servicecatalog.scan_provisioned_products_single_page_{self.account_id}_{self.region}",
            f"servicecatalog.describe_provisioning_artifact_{self.account_id}_{self.region}",
        ]

    def run(self):
        self.info(
            f"starting dry run terminate try {self.try_count} of {self.retry_count}"
        )

        role = f"arn:aws:iam::{self.account_id}:role/servicecatalog-puppet/PuppetRole"
        with betterboto_client.CrossAccountClientContextManager(
            "servicecatalog",
            role,
            f"sc-{self.region}-{self.account_id}",
            region_name=self.region,
        ) as service_catalog:
            self.info(
                f"[{self.launch_name}] {self.account_id}:{self.region} :: looking for previous failures"
            )
            r = aws.get_provisioned_product_details(
                self.product_id, self.launch_name, service_catalog
            )

            if r is None:
                self.write_result(
                    "-", "-", constants.NO_CHANGE, notes="There is nothing to terminate"
                )
            else:
                provisioned_product_name = (
                    service_catalog.describe_provisioning_artifact(
                        ProvisioningArtifactId=r.get("ProvisioningArtifactId"),
                        ProductId=self.product_id,
                    )
                    .get("ProvisioningArtifactDetail")
                    .get("Name")
                )

                if r.get("Status") != "TERMINATED":
                    self.write_result(
                        provisioned_product_name,
                        "-",
                        constants.CHANGE,
                        notes="The product would be terminated",
                    )
                else:
                    self.write_result(
                        "-",
                        "-",
                        constants.CHANGE,
                        notes="The product is already terminated",
                    )


class ResetProvisionedProductOwnerTask(ProvisioningTask):
    launch_name = luigi.Parameter()
    account_id = luigi.Parameter()
    region = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "launch_name": self.launch_name,
            "account_id": self.account_id,
            "region": self.region,
        }

    def api_calls_used(self):
        return [
            f"servicecatalog.scan_provisioned_products_single_page_{self.account_id}_{self.region}",
            f"servicecatalog.update_provisioned_product_properties_{self.account_id}_{self.region}",
        ]

    def run(self):
        logger_prefix = f"[{self.launch_name}] {self.account_id}:{self.region}"
        self.info(f"{logger_prefix} :: starting ResetProvisionedProductOwnerTask")

        with betterboto_client.CrossAccountClientContextManager(
            "servicecatalog",
            f"arn:aws:iam::{self.account_id}:role/servicecatalog-puppet/PuppetRole",
            f"sc-{self.region}-{self.account_id}",
            region_name=self.region,
        ) as service_catalog:
            self.info(
                f"[{logger_prefix} :: Checking if existing provisioned product exists"
            )
            all_results = service_catalog.scan_provisioned_products_single_page(
                AccessLevelFilter={"Key": "Account", "Value": "self"},
            ).get("ProvisionedProducts", [])
            changes_made = list()
            for result in all_results:
                if result.get("Name") == self.launch_name:
                    provisioned_product_id = result.get("Id")
                    self.info(
                        f"[{logger_prefix} :: Ensuring current provisioned product owner is correct"
                    )
                    changes_made.append(result)
                    service_catalog.update_provisioned_product_properties(
                        ProvisionedProductId=provisioned_product_id,
                        ProvisionedProductProperties={
                            "OWNER": f"arn:aws:iam::{self.account_id}:role/servicecatalog-puppet/PuppetRole"
                        },
                    )
            self.write_output(changes_made)


class RunDeployInSpokeTask(tasks.PuppetTask):
    manifest_file_path = luigi.Parameter()
    puppet_account_id = luigi.Parameter()
    account_id = luigi.Parameter()

    home_region = luigi.Parameter()
    regions = luigi.ListParameter()
    should_collect_cloudformation_events = luigi.BoolParameter()
    should_forward_events_to_eventbridge = luigi.BoolParameter()
    should_forward_failures_to_opscenter = luigi.BoolParameter()

    def params_for_results_display(self):
        return {
            "manifest_file_path": self.manifest_file_path,
            "puppet_account_id": self.puppet_account_id,
            "account_id": self.account_id,
        }

    def run(self):
        with betterboto_client.CrossAccountClientContextManager(
            "s3",
            f"arn:aws:iam::{self.puppet_account_id}:role/servicecatalog-puppet/PuppetRole",
            f"s3-{self.puppet_account_id}",
        ) as s3:
            bucket = f"sc-puppet-spoke-deploy-{self.puppet_account_id}"
            key = f"{os.getenv('CODEBUILD_BUILD_NUMBER', '0')}.yaml"
            s3.put_object(
                Body=open(self.manifest_file_path).read(), Bucket=bucket, Key=key,
            )
            signed_url = s3.generate_presigned_url(
                "get_object",
                Params={"Bucket": bucket, "Key": key},
                ExpiresIn=60 * 60 * 24,
            )
        with betterboto_client.ClientContextManager("ssm") as ssm:
            response = ssm.get_parameter(Name="service-catalog-puppet-version")
            version = response.get("Parameter").get("Value")
        with betterboto_client.CrossAccountClientContextManager(
            "codebuild",
            f"arn:aws:iam::{self.account_id}:role/servicecatalog-puppet/PuppetRole",
            f"codebuild-{self.account_id}",
        ) as codebuild:
            response = codebuild.start_build(
                projectName=constants.EXECUTION_SPOKE_CODEBUILD_PROJECT_NAME,
                environmentVariablesOverride=[
                    {"name": "VERSION", "value": version, "type": "PLAINTEXT"},
                    {"name": "MANIFEST_URL", "value": signed_url, "type": "PLAINTEXT"},
                    {
                        "name": "PUPPET_ACCOUNT_ID",
                        "value": self.puppet_account_id,
                        "type": "PLAINTEXT",
                    },
                    {
                        "name": "HOME_REGION",
                        "value": self.home_region,
                        "type": "PLAINTEXT",
                    },
                    {
                        "name": "REGIONS",
                        "value": ",".join(self.regions),
                        "type": "PLAINTEXT",
                    },
                    {
                        "name": "SHOULD_COLLECT_CLOUDFORMATION_EVENTS",
                        "value": str(self.should_collect_cloudformation_events),
                        "type": "PLAINTEXT",
                    },
                    {
                        "name": "SHOULD_FORWARD_EVENTS_TO_EVENTBRIDGE",
                        "value": str(self.should_forward_events_to_eventbridge),
                        "type": "PLAINTEXT",
                    },
                    {
                        "name": "SHOULD_FORWARD_FAILURES_TO_OPSCENTER",
                        "value": str(self.should_forward_failures_to_opscenter),
                        "type": "PLAINTEXT",
                    },
                ],
            )
        self.write_output(response)


class LaunchInSpokeTask(ProvisioningTask, manifest_tasks.ManifestMixen):
    launch_name = luigi.Parameter()
    puppet_account_id = luigi.Parameter()
    should_use_sns = luigi.BoolParameter()
    should_use_product_plans = luigi.BoolParameter()
    include_expanded_from = luigi.BoolParameter()
    single_account = luigi.Parameter()
    is_dry_run = luigi.BoolParameter()
    execution_mode = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "launch_name": self.launch_name,
        }

    def generate_tasks(self, task_defs, manifest):
        self.info("generate_provisions")
        provisions = []
        accounts_by_id = dict()
        for a in manifest.get("accounts"):
            accounts_by_id[a.get("account_id")] = a

        for task_def in task_defs:
            if self.single_account == "None" or self.single_account is None:
                pass
            else:
                if task_def.get("account_id") != self.single_account:
                    continue
            task_status = task_def.get("status")
            del task_def["status"]
            del task_def["depends_on"]
            task_def["is_dry_run"] = self.is_dry_run
            run_deploy_in_spoke_task_params = dict(
                manifest_file_path=self.manifest_file_path,
                puppet_account_id=self.puppet_account_id,
                account_id=task_def.get("account_id"),
                home_region=config.get_home_region(self.puppet_account_id),
                regions=config.get_regions(self.puppet_account_id),
                should_collect_cloudformation_events=config.get_should_use_sns(
                    self.puppet_account_id
                ),
                should_forward_events_to_eventbridge=config.get_should_use_eventbridge(
                    self.puppet_account_id
                ),
                should_forward_failures_to_opscenter=config.get_should_forward_failures_to_opscenter(
                    self.puppet_account_id
                ),
            )

            if task_status == constants.PROVISIONED:
                provisioning_parameters = {}
                for p in ProvisionProductTask.get_param_names(include_significant=True):
                    provisioning_parameters[p] = task_def.get(p)

                if self.is_dry_run:
                    provisions.append(
                        ProvisionProductDryRunTask(**provisioning_parameters)
                    )
                else:
                    provisions.append(
                        RunDeployInSpokeTask(**run_deploy_in_spoke_task_params)
                    )

            elif task_status == constants.TERMINATED:
                terminating_parameters = {}
                for p in TerminateProductTask.get_param_names(include_significant=True):
                    terminating_parameters[p] = task_def.get(p)

                if self.is_dry_run:
                    provisions.append(
                        TerminateProductDryRunTask(**terminating_parameters)
                    )
                else:
                    provisions.append(
                        RunDeployInSpokeTask(**run_deploy_in_spoke_task_params)
                    )
            else:
                raise Exception(f"Unsupported status of {task_status}")
        self.info(f"len of provisions: {len(provisions)}")
        return provisions

    def requires(self):
        configuration = manifest_utils_for_launches.get_configuration_from_launch(
            self.manifest, self.launch_name
        )
        configuration["single_account"] = self.single_account
        configuration["is_dry_run"] = self.is_dry_run
        configuration["puppet_account_id"] = self.puppet_account_id
        configuration["should_use_sns"] = self.should_use_sns
        configuration["should_use_product_plans"] = self.should_use_product_plans

        launch_tasks_def = self.manifest.get_task_defs_from_details(
            self.puppet_account_id, False, self.launch_name, configuration, "launches"
        )
        requirements = dict(
            launches=self.generate_tasks(launch_tasks_def, self.manifest)
        )
        return requirements

    def run(self):
        self.info("started")
        self.write_output(self.params_for_results_display())


class LaunchTask(ProvisioningTask, manifest_tasks.ManifestMixen):
    launch_name = luigi.Parameter()
    puppet_account_id = luigi.Parameter()
    should_use_sns = luigi.BoolParameter()
    should_use_product_plans = luigi.BoolParameter()
    include_expanded_from = luigi.BoolParameter()
    single_account = luigi.Parameter()
    is_dry_run = luigi.BoolParameter()
    execution_mode = luigi.Parameter()
    cache_invalidator = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "launch_name": self.launch_name,
            "execution_mode": self.execution_mode,
            "cache_invalidator": self.cache_invalidator,
        }

    def generate_provisions(self, task_defs, manifest):
        self.info("generate_provisions")
        provisions = []
        accounts_by_id = dict()
        for a in manifest.get("accounts"):
            accounts_by_id[a.get("account_id")] = a

        version_details = self.input().get("version_details")

        for task_def in task_defs:
            if self.single_account == "None" or self.single_account is None:
                pass
            else:
                if task_def.get("account_id") != self.single_account:
                    continue
            task_status = task_def.get("status")
            del task_def["status"]
            del task_def["depends_on"]
            task_def["is_dry_run"] = self.is_dry_run

            puppet_account_id = task_def.get("puppet_account_id")
            portfolio = task_def.get("portfolio")
            product = task_def.get("product")
            version = task_def.get("version")
            account_id = task_def.get("account_id")
            region = task_def.get("region")

            d = json.loads(
                version_details[
                    "_".join(
                        [
                            str(puppet_account_id),
                            portfolio,
                            product,
                            version,
                            str(account_id),
                            region,
                        ]
                    )
                ]
                .open("r")
                .read()
            )

            version_id = d.get("version_details").get("version_id")

            product_id = d.get("product_details").get("product_id")

            portfolio_id = d.get("portfolio_details").get("portfolio_id")

            if task_status == constants.PROVISIONED:
                provisioning_parameters = dict()

                for p in ProvisionProductTask.get_param_names(include_significant=True):
                    provisioning_parameters[p] = task_def.get(p)

                provisioning_parameters.update(
                    dict(
                        version_id=version_id,
                        product_id=product_id,
                        portfolio_id=portfolio_id,
                        cache_invalidator=self.cache_invalidator,
                        manifest_file_path=self.manifest_file_path,
                    )
                )

                if self.is_dry_run:
                    provisions.append(
                        ProvisionProductDryRunTask(**provisioning_parameters)
                    )
                else:
                    provisions.append(ProvisionProductTask(**provisioning_parameters))

            elif task_status == constants.TERMINATED:
                terminating_parameters = dict()
                for p in TerminateProductTask.get_param_names(include_significant=True):
                    terminating_parameters[p] = task_def.get(p)

                    terminating_parameters.update(
                        dict(
                            version_id=version_id,
                            product_id=product_id,
                            portfolio_id=portfolio_id,
                            cache_invalidator=self.cache_invalidator,
                            manifest_file_path=self.manifest_file_path,
                        )
                    )

                if self.is_dry_run:
                    provisions.append(
                        TerminateProductDryRunTask(**terminating_parameters)
                    )
                else:
                    provisions.append(TerminateProductTask(**terminating_parameters))
            else:
                raise Exception(f"Unsupported status of {task_status}")
        return provisions

    def requires(self):
        launch = self.manifest.get("launches", {}).get(self.launch_name)

        if not launch:
            raise Exception(self.launch_name)

        dependencies = list()
        version_details = dict()
        requirements = dict(
            dependencies=dependencies,
            version_details=version_details,
            generate_shares=generate_tasks.GenerateSharesTask(
                puppet_account_id=self.puppet_account_id,
                manifest_file_path=self.manifest_file_path,
                should_use_sns=self.should_use_sns,
                section=constants.LAUNCHES,
                cache_invalidator=self.cache_invalidator,
            ),
        )

        for dependency in launch.get("depends_on", []):
            if isinstance(dependency, str):
                dependencies.append(
                    self.__class__(
                        launch_name=dependency,
                        manifest_file_path=self.manifest_file_path,
                        puppet_account_id=self.puppet_account_id,
                        should_use_sns=self.should_use_sns,
                        should_use_product_plans=self.should_use_product_plans,
                        include_expanded_from=self.include_expanded_from,
                        single_account=self.single_account,
                        is_dry_run=self.is_dry_run,
                        execution_mode=self.execution_mode,
                        cache_invalidator=self.cache_invalidator,
                    )
                )
            else:
                dependency_type = dependency.get("type", "launch")
                if dependency_type == "launch":
                    dependencies.append(
                        self.__class__(
                            launch_name=dependency,
                            manifest_file_path=self.manifest_file_path,
                            puppet_account_id=self.puppet_account_id,
                            should_use_sns=self.should_use_sns,
                            should_use_product_plans=self.should_use_product_plans,
                            include_expanded_from=self.include_expanded_from,
                            single_account=self.single_account,
                            is_dry_run=self.is_dry_run,
                            execution_mode=self.execution_mode,
                            cache_invalidator=self.cache_invalidator,
                        )
                    )
                elif dependency_type == "lambda-invocation":
                    from servicecatalog_puppet.workflow.lambda_invocations import (
                        LambdaInvocationTask,
                    )

                    dependencies.append(
                        LambdaInvocationTask(
                            lambda_invocation_name=dependency.get("name"),
                            manifest_file_path=self.manifest_file_path,
                            puppet_account_id=self.puppet_account_id,
                            should_use_sns=self.should_use_sns,
                            should_use_product_plans=self.should_use_product_plans,
                            include_expanded_from=self.include_expanded_from,
                            single_account=self.single_account,
                            is_dry_run=self.is_dry_run,
                            cache_invalidator=self.cache_invalidator,
                        )
                    )

        launch_tasks_defs = self.get_launch_tasks_defs()
        for task_def in launch_tasks_defs:
            if self.single_account == "None" or self.single_account is None:
                pass
            else:
                if task_def.get("account_id") != self.single_account:
                    continue

            puppet_account_id = task_def.get("puppet_account_id")
            portfolio = task_def.get("portfolio")
            product = task_def.get("product")
            version = task_def.get("version")
            account_id = task_def.get("account_id")
            region = task_def.get("region")

            version_details[
                "_".join(
                    [
                        str(puppet_account_id),
                        portfolio,
                        product,
                        version,
                        str(account_id),
                        region,
                    ]
                )
            ] = portfoliomanagement_tasks.GetVersionDetailsByNames(
                manifest_file_path=self.manifest_file_path,
                puppet_account_id=puppet_account_id,
                portfolio=portfolio,
                product=product,
                version=version,
                account_id=puppet_account_id,
                region=region,
                cache_invalidator=self.cache_invalidator,
            )

        return requirements

    @lru_cache()
    def get_launch_tasks_defs(self):
        launch = self.manifest.get("launches").get(self.launch_name)
        if not launch:
            raise Exception(self.launch_name)

        configuration = manifest_utils_for_launches.get_configuration_from_launch(
            self.manifest, self.launch_name
        )
        configuration["single_account"] = self.single_account
        configuration["is_dry_run"] = self.is_dry_run
        configuration["puppet_account_id"] = self.puppet_account_id
        configuration["should_use_sns"] = self.should_use_sns
        configuration["should_use_product_plans"] = self.should_use_product_plans

        configuration["execution"] = launch.get("execution")

        launch_tasks_def = self.manifest.get_task_defs_from_details(
            self.puppet_account_id, False, self.launch_name, configuration, "launches"
        )
        return launch_tasks_def

    def run(self):
        self.info("started")
        launch_tasks_def = self.get_launch_tasks_defs()

        self.info(f"{self.uid} starting pre actions")
        pre_actions = self.manifest.get_actions_from(
            self.launch_name, "pre", "launches"
        )
        yield [
            portfoliomanagement_tasks.ProvisionActionTask(
                self.manifest_file_path,
                **p,
                puppet_account_id=self.puppet_account_id,
                cache_invalidator=self.cache_invalidator,
            )
            for p in pre_actions
        ]
        self.info(f"{self.uid} finished pre actions")

        self.info(f"{self.uid} starting launches")
        ls = self.generate_provisions(launch_tasks_def, self.manifest)
        yield ls
        self.info(f"{self.uid} finished launches")

        self.info(f"{self.uid} starting post actions")
        post_actions = self.manifest.get_actions_from(
            self.launch_name, "post", "launches"
        )
        yield [
            portfoliomanagement_tasks.ProvisionActionTask(
                self.manifest_file_path,
                **p,
                puppet_account_id=self.puppet_account_id,
                cache_invalidator=self.cache_invalidator,
            )
            for p in post_actions
        ]
        self.info(f"{self.uid} finished post actions")

        self.write_output(dict(**self.params_for_results_display(), skipped=False))
        self.info("Finished")


class SpokeLocalPortfolioTask(ProvisioningTask, manifest_tasks.ManifestMixen):
    spoke_local_portfolio_name = luigi.Parameter()
    puppet_account_id = luigi.Parameter()
    should_use_sns = luigi.BoolParameter()
    should_use_product_plans = luigi.BoolParameter()
    include_expanded_from = luigi.BoolParameter()
    single_account = luigi.Parameter()
    is_dry_run = luigi.BoolParameter()
    depends_on = luigi.ListParameter()
    sharing_mode = luigi.Parameter()
    cache_invalidator = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "spoke_local_portfolio_name": self.spoke_local_portfolio_name,
            "sharing_mode": self.sharing_mode,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        self.info("requires")
        dependencies = [
            LaunchTask(
                launch_name=dependency,
                manifest_file_path=self.manifest_file_path,
                puppet_account_id=self.puppet_account_id,
                should_use_sns=self.should_use_sns,
                should_use_product_plans=self.should_use_product_plans,
                include_expanded_from=self.include_expanded_from,
                single_account=self.single_account,
                is_dry_run=self.is_dry_run,
                execution_mode="hub",
                cache_invalidator=self.cache_invalidator,
            )
            for dependency in self.depends_on
        ]
        self.info(f"dependencies are {dependencies}")

        task_defs = self.get_task_defs()
        portfolio_ids = dict()

        self.info("requires:about to iter")
        for task_def in task_defs:
            # if task_def.get("status") == constants.SPOKE_LOCAL_PORTFOLIO_STATUS_SHARED:  # Dodgy if
            portfolio_ids[
                "_".join(
                    [
                        str(self.puppet_account_id),
                        task_def.get("portfolio"),
                        str(task_def.get("account_id")),
                        task_def.get("region"),
                    ]
                )
            ] = portfoliomanagement_tasks.GetPortfolioByPortfolioName(
                manifest_file_path=self.manifest_file_path,
                puppet_account_id=self.puppet_account_id,
                portfolio=task_def.get("portfolio"),
                account_id=self.puppet_account_id,
                region=task_def.get("region"),
                cache_invalidator=self.cache_invalidator,
            )

        self.info("requires:about to return")
        return dict(dependencies=dependencies, portfolio_ids=portfolio_ids,)

    def generate_tasks(self, task_defs):
        self.info("generate_tasks in")
        if len(task_defs) == 0:
            logger.warning(
                f"The configuration for this share does not include any target accounts: {self.spoke_local_portfolio_name}"
            )
            return []
        first_task_def = task_defs[0]
        # portfolio = first_task_def.get("portfolio")
        tasks = []
        portfolio_ids = self.input().get("portfolio_ids")

        for task_def in task_defs:
            self.info("generate_tasks main loop iteration 1")
            p = (
                portfolio_ids[
                    "_".join(
                        [
                            str(self.puppet_account_id),
                            task_def.get("portfolio"),
                            str(task_def.get("account_id")),
                            task_def.get("region"),
                        ]
                    )
                ]
                .open("r")
                .read()
            )
            p = json.loads(p)
            portfolio_id = p.get("portfolio_id")

            product_generation_method = task_def.get(
                "product_generation_method", "copy"
            )

            sharing_mode = task_def.get("sharing_mode", constants.SHARING_MODE_DEFAULT)

            self.info("generate_tasks main loop iteration 2")
            if (
                task_def.get("status")
                == constants.SPOKE_LOCAL_PORTFOLIO_STATUS_TERMINATED
            ):
                tasks.append(
                    portfoliomanagement_tasks.DeletePortfolio(
                        manifest_file_path=self.manifest_file_path,
                        account_id=task_def.get("account_id"),
                        region=task_def.get("region"),
                        portfolio=task_def.get("portfolio"),
                        product_generation_method=product_generation_method,
                        puppet_account_id=task_def.get("puppet_account_id"),
                    )
                )
            elif (
                task_def.get("status") == constants.SPOKE_LOCAL_PORTFOLIO_STATUS_SHARED
            ):

                create_spoke_local_portfolio_task_params = dict(
                    manifest_file_path=self.manifest_file_path,
                    puppet_account_id=self.puppet_account_id,
                    account_id=task_def.get("account_id"),
                    region=task_def.get("region"),
                    portfolio=task_def.get("portfolio"),
                    organization=task_def.get("organization"),
                    portfolio_id=portfolio_id,
                    sharing_mode=sharing_mode,
                    cache_invalidator=self.cache_invalidator,
                )

                create_spoke_local_portfolio_task = portfoliomanagement_tasks.CreateSpokeLocalPortfolioTask(
                    **create_spoke_local_portfolio_task_params
                )
                tasks.append(create_spoke_local_portfolio_task)

            create_spoke_local_portfolio_task_as_dependency_params = dict(
                manifest_file_path=self.manifest_file_path,
                account_id=task_def.get("account_id"),
                region=task_def.get("region"),
                portfolio=task_def.get("portfolio"),
                organization=task_def.get("organization"),
                portfolio_id=portfolio_id,
                sharing_mode=sharing_mode,
                cache_invalidator=self.cache_invalidator,
            )

            if len(task_def.get("associations", [])) > 0:
                create_associations_for_portfolio_task = portfoliomanagement_tasks.CreateAssociationsForPortfolioTask(
                    **create_spoke_local_portfolio_task_as_dependency_params,
                    associations=task_def.get("associations"),
                    puppet_account_id=task_def.get("puppet_account_id"),
                    should_use_sns=task_def.get("should_use_sns"),
                )
                tasks.append(create_associations_for_portfolio_task)

            launch_constraints = task_def.get("constraints", {}).get("launch", [])

            if product_generation_method == "import":
                import_into_spoke_local_portfolio_task = portfoliomanagement_tasks.ImportIntoSpokeLocalPortfolioTask(
                    **create_spoke_local_portfolio_task_as_dependency_params,
                    puppet_account_id=task_def.get("puppet_account_id"),
                )
                tasks.append(import_into_spoke_local_portfolio_task)
            else:
                copy_into_spoke_local_portfolio_task = portfoliomanagement_tasks.CopyIntoSpokeLocalPortfolioTask(
                    **create_spoke_local_portfolio_task_as_dependency_params,
                    puppet_account_id=task_def.get("puppet_account_id"),
                )
                tasks.append(copy_into_spoke_local_portfolio_task)

            if len(launch_constraints) > 0:
                create_launch_role_constraints_for_portfolio_task_params = {
                    "launch_constraints": launch_constraints,
                    "puppet_account_id": task_def.get("puppet_account_id"),
                    "should_use_sns": task_def.get("should_use_sns"),
                }
                create_launch_role_constraints_for_portfolio = portfoliomanagement_tasks.CreateLaunchRoleConstraintsForPortfolio(
                    **create_spoke_local_portfolio_task_as_dependency_params,
                    **create_launch_role_constraints_for_portfolio_task_params,
                    product_generation_method=product_generation_method,
                )
                tasks.append(create_launch_role_constraints_for_portfolio)
        self.info(f"tasks len are {len(tasks)}")
        self.info("generate_tasks out")
        return tasks

    @lru_cache()
    def get_task_defs(self):
        self.info("get_task_defs in")
        spoke_local_portfolio_details = self.manifest.get("spoke-local-portfolios").get(
            self.spoke_local_portfolio_name
        )
        configuration = {
            "spoke_local_portfolio_name": self.spoke_local_portfolio_name,
            "status": spoke_local_portfolio_details.get("status", "shared"),
            "portfolio": spoke_local_portfolio_details.get("portfolio"),
            "associations": spoke_local_portfolio_details.get("associations", []),
            "constraints": spoke_local_portfolio_details.get("constraints", {}),
            "depends_on": spoke_local_portfolio_details.get("depends_on", []),
            "retry_count": spoke_local_portfolio_details.get("retry_count", 1),
            "requested_priority": spoke_local_portfolio_details.get(
                "requested_priority", 0
            ),
            "worker_timeout": spoke_local_portfolio_details.get(
                "timeoutInSeconds", constants.DEFAULT_TIMEOUT
            ),
            "product_generation_method": spoke_local_portfolio_details.get(
                "product_generation_method", "copy"
            ),
            "single_account": self.single_account,
            "is_dry_run": self.is_dry_run,
            "puppet_account_id": self.puppet_account_id,
            "should_use_sns": self.should_use_sns,
            "should_use_product_plans": self.should_use_product_plans,
        }
        configuration.update(
            manifest_utils.get_configuration_overrides(
                self.manifest, spoke_local_portfolio_details
            )
        )

        self.info("get_task_defs out")

        return self.manifest.get_task_defs_from_details(
            self.puppet_account_id,
            True,
            self.spoke_local_portfolio_name,
            configuration,
            "spoke-local-portfolios",
        )

    def run(self):
        self.info("started")

        task_defs = self.get_task_defs()

        self.info("about to get the pre actions")
        pre_actions = self.manifest.get_actions_from(
            self.spoke_local_portfolio_name, "pre", "spoke-local-portfolios"
        )
        self.info("about to get the post actions")
        post_actions = self.manifest.get_actions_from(
            self.spoke_local_portfolio_name, "post", "spoke-local-portfolios"
        )

        self.info(f"starting pre actions")
        yield [
            portfoliomanagement_tasks.ProvisionActionTask(
                **p,
                puppet_account_id=self.puppet_account_id,
                cache_invalidator=self.cache_invalidator,
            )
            for p in pre_actions
        ]
        self.info(f"finished pre actions")

        self.info(f"starting launches")
        yield self.generate_tasks(task_defs)
        self.info(f"{self.uid} finished launches")

        self.info(f"{self.uid} starting post actions")
        yield [
            portfoliomanagement_tasks.ProvisionActionTask(
                **p,
                puppet_account_id=self.puppet_account_id,
                cache_invalidator=self.cache_invalidator,
            )
            for p in post_actions
        ]
        self.info(f"{self.uid} finished post actions")

        self.write_output(self.params_for_results_display())
        self.info("finished")



from . import tasks_unit_tests


class GetVersionIdByVersionNameTest(tasks_unit_tests.PuppetTaskUnitTest):

    puppet_account_id = "01234567890"
    manifest_file_path = "tcvyuiho"
    portfolio = "port1"
    portfolio_id = "psdsdort1"
    product = "prod1"
    product_id = "sdsdprod1"
    version = "v1"
    account_id = "23089479278643892"
    region = "eu-west-1"
    cache_invalidator = "foo"

    def setUp(self) -> None:
        from . import portfoliomanagement

        self.sut = portfoliomanagement.GetVersionIdByVersionName(
            manifest_file_path=self.manifest_file_path,
            puppet_account_id=self.puppet_account_id,
            portfolio=self.portfolio,
            portfolio_id=self.portfolio_id,
            product=self.product,
            product_id=self.product_id,
            version=self.version,
            account_id=self.account_id,
            region=self.region,
            cache_invalidator=self.cache_invalidator,
        )

    def test_params_for_results_display(self):
        expected_result = {
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "portfolio_id": self.portfolio_id,
            "region": self.region,
            "product": self.product,
            "product_id": self.product_id,
            "version": self.version,
            "account_id": self.account_id,
            "cache_invalidator": self.cache_invalidator,
        }
        self.assertEqual(expected_result, self.sut.params_for_results_display())


class SearchProductsAsAdminTaskTest(tasks_unit_tests.PuppetTaskUnitTest):
    puppet_account_id = "01234567890"
    manifest_file_path = "tcvyuiho"
    portfolio = "port1"
    portfolio_id = "psdlsdm;lmort1"
    account_id = "23089479278643892"
    region = "eu-west-1"
    cache_invalidator = "foo"

    def setUp(self) -> None:
        from . import portfoliomanagement

        self.sut = portfoliomanagement.SearchProductsAsAdminTask(
            puppet_account_id=self.puppet_account_id,
            manifest_file_path=self.manifest_file_path,
            portfolio=self.portfolio,
            portfolio_id=self.portfolio_id,
            account_id=self.account_id,
            region=self.region,
            cache_invalidator=self.cache_invalidator,
        )

    def test_params_for_results_display(self):
        expected_result = {
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "portfolio_id": self.portfolio_id,
            "region": self.region,
            "account_id": self.account_id,
            "cache_invalidator": self.cache_invalidator,
        }
        self.assertEqual(expected_result, self.sut.params_for_results_display())


class GetProductIdByProductNameTest(tasks_unit_tests.PuppetTaskUnitTest):
    puppet_account_id = "01234567890"
    manifest_file_path = "tcvyuiho"
    portfolio = "portfolio1"
    portfolio_id = "sdsdsdsdportfolio1"
    product = "prod1"
    account_id = "23089479278643892"
    region = "eu-west-1"
    cache_invalidator = "foo"

    def setUp(self) -> None:
        from . import portfoliomanagement

        self.sut = portfoliomanagement.GetProductIdByProductName(
            puppet_account_id=self.puppet_account_id,
            manifest_file_path=self.manifest_file_path,
            portfolio=self.portfolio,
            portfolio_id=self.portfolio_id,
            product=self.product,
            account_id=self.account_id,
            region=self.region,
            cache_invalidator=self.cache_invalidator,
        )

    def test_params_for_results_display(self):
        expected_result = {
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "portfolio_id": self.portfolio_id,
            "region": self.region,
            "product": self.product,
            "account_id": self.account_id,
            "cache_invalidator": self.cache_invalidator,
        }
        self.assertEqual(expected_result, self.sut.params_for_results_display())

    def test_requires(self):
        from . import portfoliomanagement

        expected_result = {
            "search_products_as_admin": portfoliomanagement.SearchProductsAsAdminTask(
                manifest_file_path=self.manifest_file_path,
                puppet_account_id=self.puppet_account_id,
                portfolio=self.portfolio,
                portfolio_id=self.portfolio_id,
                account_id=self.account_id,
                region=self.region,
                cache_invalidator=self.cache_invalidator,
            ),
        }
        self.assertEqual(expected_result, self.sut.requires())


class GetPortfolioByPortfolioNameTest(tasks_unit_tests.PuppetTaskUnitTest):
    puppet_account_id = "01234567890"
    manifest_file_path = "tcvyuiho"
    portfolio = "portfolio1"
    account_id = "23089479278643892"
    region = "eu-west-1"
    cache_invalidator = "foo"

    def setUp(self) -> None:
        from . import portfoliomanagement

        self.sut = portfoliomanagement.GetPortfolioByPortfolioName(
            puppet_account_id=self.puppet_account_id,
            manifest_file_path=self.manifest_file_path,
            portfolio=self.portfolio,
            account_id=self.account_id,
            region=self.region,
            cache_invalidator=self.cache_invalidator,
        )

    def test_params_for_results_display(self):
        expected_result = {
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "region": self.region,
            "account_id": self.account_id,
            "cache_invalidator": self.cache_invalidator,
        }
        self.assertEqual(expected_result, self.sut.params_for_results_display())


# class ProvisionActionTaskTest(tasks_unit_tests.PuppetTaskUnitTest):
#
#     def setUp(self) -> None:
#         from . import portfoliomanagement
#
#         self.sut = portfoliomanagement.ProvisionActionTask(
#         )
#
#     def test_params_for_results_display(self):
#         expected_result = {
#             "type": self.type,
#             "source": self.source,
#             "phase": self.phase,
#             "source_type": self.source_type,
#             "name": self.name,
#             "project_name": self.project_name,
#             "account_id": self.account_id,
#             "region": self.region,
#         }
#         self.assertEqual(expected_result, self.sut.params_for_results_display())
#
#
# class CreateSpokeLocalPortfolioTaskTest(tasks_unit_tests.PuppetTaskUnitTest):
#
#     def setUp(self) -> None:
#         from . import portfoliomanagement
#
#         self.sut = portfoliomanagement.CreateSpokeLocalPortfolioTask(
#         )
#
#     def test_params_for_results_display(self):
#         expected_result = {
#             "account_id": self.account_id,
#             "region": self.region,
#             "portfolio": self.portfolio,
#         }
#         self.assertEqual(expected_result, self.sut.params_for_results_display())
#
#
# class CreateAssociationsForPortfolioTaskTest(tasks_unit_tests.PuppetTaskUnitTest):
#
#     def setUp(self) -> None:
#         from . import portfoliomanagement
#
#         self.sut = portfoliomanagement.CreateAssociationsForPortfolioTask(
#         )
#
#     def test_requires(self):
#         expected_result = {
#             "create_spoke_local_portfolio_task": CreateSpokeLocalPortfolioTask(
#                 manifest_file_path=self.manifest_file_path,
#                 puppet_account_id=self.puppet_account_id,
#                 account_id=self.account_id,
#                 region=self.region,
#                 portfolio=self.portfolio,
#                 organization=self.organization,
#             ),
#         }
#         self.assertEqual(expected_result, self.sut.requires())
#
#     def test_params_for_results_display(self):
#         expected_result = {
#             "account_id": self.account_id,
#             "region": self.region,
#             "portfolio": self.portfolio,
#         }
#         self.assertEqual(expected_result, self.sut.params_for_results_display())
#
#
# class GetProductsAndProvisioningArtifactsTaskTest(tasks_unit_tests.PuppetTaskUnitTest):
#
#     def setUp(self) -> None:
#         from . import portfoliomanagement
#
#         self.sut = portfoliomanagement.GetProductsAndProvisioningArtifactsTask(
#         )
#
#     def test_params_for_results_display(self):
#         expected_result = {
#             "region": self.region,
#             "portfolio": self.portfolio,
#             "puppet_account_id": self.puppet_account_id,
#         }
#         self.assertEqual(expected_result, self.sut.params_for_results_display())
#
#     def test_requires(self):
#         expected_result = {
#             "search_products_as_admin": SearchProductsAsAdminTask(
#                 manifest_file_path=self.manifest_file_path,
#                 puppet_account_id=self.puppet_account_id,
#                 portfolio=self.portfolio,
#                 region=self.region,
#                 account_id=self.puppet_account_id,
#             )
#         }
#         self.assertEqual(expected_result, self.sut.requires())
#
#
# class CopyIntoSpokeLocalPortfolioTaskTest(tasks_unit_tests.PuppetTaskUnitTest):
#
#     def setUp(self) -> None:
#         from . import portfoliomanagement
#
#         self.sut = portfoliomanagement.CopyIntoSpokeLocalPortfolioTask(
#         )
#
#     def test_requires(self):
#         expected_result = {
#             "create_spoke_local_portfolio": CreateSpokeLocalPortfolioTask(
#                 manifest_file_path=self.manifest_file_path,
#                 account_id=self.account_id,
#                 region=self.region,
#                 portfolio=self.portfolio,
#                 organization=self.organization,
#                 puppet_account_id=self.puppet_account_id,
#             ),
#             "products_and_provisioning_artifacts": GetProductsAndProvisioningArtifactsTask(
#                 manifest_file_path=self.manifest_file_path,
#                 region=self.region,
#                 portfolio=self.portfolio,
#                 puppet_account_id=self.puppet_account_id,
#             ),
#         }
#         self.assertEqual(expected_result, self.sut.requires())
#
#     def test_params_for_results_display(self):
#         expected_result = {
#             "account_id": self.account_id,
#             "region": self.region,
#             "portfolio": self.portfolio,
#         }
#         self.assertEqual(expected_result, self.sut.params_for_results_display())
#
#
# class ImportIntoSpokeLocalPortfolioTaskTest(tasks_unit_tests.PuppetTaskUnitTest):
#
#     def setUp(self) -> None:
#         from . import portfoliomanagement
#
#         self.sut = portfoliomanagement.ImportIntoSpokeLocalPortfolioTask(
#         )
#
#     def test_requires(self):
#         expected_result = {
#             "create_spoke_local_portfolio": CreateSpokeLocalPortfolioTask(
#                 manifest_file_path=self.manifest_file_path,
#                 puppet_account_id=self.puppet_account_id,
#                 account_id=self.account_id,
#                 region=self.region,
#                 portfolio=self.portfolio,
#                 organization=self.organization,
#             ),
#             "products_and_provisioning_artifacts": GetProductsAndProvisioningArtifactsTask(
#                 manifest_file_path=self.manifest_file_path,
#                 region=self.region,
#                 portfolio=self.portfolio,
#                 puppet_account_id=self.puppet_account_id,
#             ),
#             "hub_portfolio": GetPortfolioByPortfolioName(
#                 manifest_file_path=self.manifest_file_path,
#                 puppet_account_id=self.puppet_account_id,
#                 portfolio=self.portfolio,
#                 account_id=self.puppet_account_id,
#                 region=self.region,
#             ),
#         }
#         self.assertEqual(expected_result, self.sut.requires())
#
#     def test_params_for_results_display(self):
#         expected_result = {
#             "account_id": self.account_id,
#             "region": self.region,
#             "portfolio": self.portfolio,
#         }
#         self.assertEqual(expected_result, self.sut.params_for_results_display())
#
#
# class CreateLaunchRoleConstraintsForPortfolioTest(tasks_unit_tests.PuppetTaskUnitTest):
#
#     def setUp(self) -> None:
#         from . import portfoliomanagement
#
#         self.sut = portfoliomanagement.CreateLaunchRoleConstraintsForPortfolio(
#         )
#
#     def test_params_for_results_display(self):
#         expected_result = {
#             "account_id": self.account_id,
#             "region": self.region,
#             "portfolio": self.portfolio,
#         }
#         self.assertEqual(expected_result, self.sut.params_for_results_display())
#
#
# class SharePortfolioTaskTest(tasks_unit_tests.PuppetTaskUnitTest):
#
#     def setUp(self) -> None:
#         from . import portfoliomanagement
#
#         self.sut = portfoliomanagement.SharePortfolioTask(
#         )
#
#     def test_params_for_results_display(self):
#         expected_result = {
#             "account_id": self.account_id,
#             "region": self.region,
#             "portfolio": self.portfolio,
#         }
#         self.assertEqual(expected_result, self.sut.params_for_results_display())
#
#
# class ShareAndAcceptPortfolioTaskTest(tasks_unit_tests.PuppetTaskUnitTest):
#
#     def setUp(self) -> None:
#         from . import portfoliomanagement
#
#         self.sut = portfoliomanagement.ShareAndAcceptPortfolioTask(
#         )
#
#     def test_params_for_results_display(self):
#         expected_result = {
#             "account_id": self.account_id,
#             "region": self.region,
#             "portfolio": self.portfolio,
#             "puppet_account_id": self.puppet_account_id,
#         }
#         self.assertEqual(expected_result, self.sut.params_for_results_display())
#
#
# class CreateAssociationsInPythonForPortfolioTaskTest(tasks_unit_tests.PuppetTaskUnitTest):
#
#     def setUp(self) -> None:
#         from . import portfoliomanagement
#
#         self.sut = portfoliomanagement.CreateAssociationsInPythonForPortfolioTask(
#         )
#
#     def test_params_for_results_display(self):
#         expected_result = {
#             "account_id": self.account_id,
#             "region": self.region,
#             "portfolio": self.portfolio,
#         }
#         self.assertEqual(expected_result, self.sut.params_for_results_display())
#
#     def test_api_calls_used(self):
#         expected_result = {
#             f"servicecatalog.associate_principal_with_portfolio_{self.region}": 1,
#         }
#         self.assertEqual(expected_result, self.sut.api_calls_used())
#
#
# class CreateShareForAccountLaunchRegionTest(tasks_unit_tests.PuppetTaskUnitTest):
#
#     def setUp(self) -> None:
#         from . import portfoliomanagement
#
#         self.sut = portfoliomanagement.CreateShareForAccountLaunchRegion(
#         )
#
#     def test_params_for_results_display(self):
#         expected_result = {
#             "account_id": self.account_id,
#             "region": self.region,
#             "portfolio": self.portfolio,
#         }
#         self.assertEqual(expected_result, self.sut.params_for_results_display())
#
#
# class DisassociateProductFromPortfolioTest(tasks_unit_tests.PuppetTaskUnitTest):
#
#     def setUp(self) -> None:
#         from . import portfoliomanagement
#
#         self.sut = portfoliomanagement.DisassociateProductFromPortfolio(
#         )
#
#     def test_params_for_results_display(self):
#         expected_result = {
#             "account_id": self.account_id,
#             "region": self.region,
#             "portfolio_id": self.portfolio_id,
#             "product_id": self.product_id,
#         }
#         self.assertEqual(expected_result, self.sut.params_for_results_display())
#
#     def test_api_calls_used(self):
#         expected_result = {
#             f"servicecatalog.disassociate_product_from_portfolio_{self.account_id}_{self.region}_{self.portfolio_id}_{self.product_id}": 1,
#         }
#         self.assertEqual(expected_result, self.sut.api_calls_used())
#
#
# class DisassociateProductsFromPortfolioTest(tasks_unit_tests.PuppetTaskUnitTest):
#
#     def setUp(self) -> None:
#         from . import portfoliomanagement
#
#         self.sut = portfoliomanagement.DisassociateProductsFromPortfolio(
#         )
#
#     def test_params_for_results_display(self):
#         expected_result = {
#             "account_id": self.account_id,
#             "region": self.region,
#             "portfolio_id": self.portfolio_id,
#         }
#         self.assertEqual(expected_result, self.sut.params_for_results_display())
#
#     def test_api_calls_used(self):
#         expected_result = {
#             f"servicecatalog.search_products_as_admin_single_page_{self.account_id}_{self.region}_{self.portfolio_id}": 1,
#         }
#         self.assertEqual(expected_result, self.sut.api_calls_used())
#
#
# class DeleteLocalPortfolioTest(tasks_unit_tests.PuppetTaskUnitTest):
#
#     def setUp(self) -> None:
#         from . import portfoliomanagement
#
#         self.sut = portfoliomanagement.DeleteLocalPortfolio(
#         )
#
#     def test_params_for_results_display(self):
#         expected_result = {
#             "account_id": self.account_id,
#             "region": self.region,
#             "portfolio_id": self.portfolio_id,
#         }
#         self.assertEqual(expected_result, self.sut.params_for_results_display())
#
#     def test_api_calls_used(self):
#         expected_result = {
#             f"servicecatalog.delete_portfolio_{self.account_id}_{self.region}_{self.portfolio_id}": 1,
#         }
#         self.assertEqual(expected_result, self.sut.api_calls_used())
#
#
# class DeletePortfolioShareTest(tasks_unit_tests.PuppetTaskUnitTest):
#
#     def setUp(self) -> None:
#         from . import portfoliomanagement
#
#         self.sut = portfoliomanagement.DeletePortfolioShare(
#         )
#
#     def test_params_for_results_display(self):
#         expected_result = {
#             "account_id": self.account_id,
#             "region": self.region,
#             "portfolio": self.portfolio,
#         }
#         self.assertEqual(expected_result, self.sut.params_for_results_display())
#
#     def test_api_calls_used(self):
#         expected_result = {
#             f"servicecatalog.list_accepted_portfolio_shares_{self.account_id}_{self.region}_{self.portfolio}": 1,
#             f"servicecatalog.delete_portfolio_share_{self.puppet_account_id}_{self.region}_{self.portfolio}": 1,
#         }
#         self.assertEqual(expected_result, self.sut.api_calls_used())
#
#
# class DeletePortfolioTest(tasks_unit_tests.PuppetTaskUnitTest):
#
#     def setUp(self) -> None:
#         from . import portfoliomanagement
#
#         self.sut = portfoliomanagement.DeletePortfolio(
#         )
#
#     def test_params_for_results_display(self):
#         expected_result = {
#             "account_id": self.account_id,
#             "region": self.region,
#             "portfolio": self.portfolio,
#         }
#         self.assertEqual(expected_result, self.sut.params_for_results_display())



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from unittest.mock import call

from pytest import fixture
import os


@fixture
def sut():
    from . import asset_helpers

    return asset_helpers


def test_resolve_from_site_packages(sut):
    # setup
    expected_result = os.path.sep.join(
        [os.path.dirname(os.path.abspath(__file__)), "foo"]
    )

    # exercise
    actual_result = sut.resolve_from_site_packages("foo")

    # verify
    assert actual_result == expected_result


def test_read_from_site_packages(sut, mocker):
    # setup
    mocked_open = mocker.patch.object(sut, "open")
    expected_param = os.path.sep.join(
        [os.path.dirname(os.path.abspath(__file__)), "foo"]
    )

    # exercise
    sut.read_from_site_packages("foo")

    # verify
    assert mocked_open.call_count == 1
    assert mocked_open.call_args == call(expected_param, "r")
    # assert mocked_open.assert_called_once_with(expected_param)



import json

import luigi

from servicecatalog_puppet.workflow import tasks as workflow_tasks
from servicecatalog_puppet.workflow import manifest as manifest_tasks
from servicecatalog_puppet.workflow.provisioning import LaunchTask
from servicecatalog_puppet.workflow import tasks
from servicecatalog_puppet import config
from betterboto import client as betterboto_client


class InvokeLambdaTask(workflow_tasks.PuppetTask):
    lambda_invocation_name = luigi.Parameter()
    region = luigi.Parameter()
    account_id = luigi.Parameter()

    function_name = luigi.Parameter()
    qualifier = luigi.Parameter()
    invocation_type = luigi.Parameter()

    puppet_account_id = luigi.Parameter()

    parameters = luigi.DictParameter()

    launch_parameters = luigi.DictParameter()
    manifest_parameters = luigi.DictParameter()
    account_parameters = luigi.DictParameter()

    all_params = []

    manifest_file_path = luigi.Parameter()
    should_use_sns = luigi.BoolParameter()
    should_use_product_plans = luigi.BoolParameter()
    include_expanded_from = luigi.BoolParameter()
    single_account = luigi.Parameter()
    is_dry_run = luigi.BoolParameter()

    cache_invalidator = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "lambda_invocation_name": self.lambda_invocation_name,
            "account_id": self.account_id,
            "region": self.region,
            "function_name": self.function_name,
            "qualifier": self.qualifier,
            "invocation_type": self.invocation_type,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        all_params = dict()
        all_params.update(self.manifest_parameters)
        all_params.update(self.launch_parameters)
        all_params.update(self.account_parameters)

        ssm_params = dict()
        for param_name, param_details in all_params.items():
            if param_details.get("ssm"):
                if param_details.get("default"):
                    del param_details["default"]
                ssm_params[param_name] = tasks.GetSSMParamTask(
                    parameter_name=param_name,
                    name=param_details.get("ssm").get("name"),
                    region=param_details.get("ssm").get(
                        "region", config.get_home_region(self.puppet_account_id)
                    ),
                    cache_invalidator=self.cache_invalidator,
                )
        self.all_params = all_params

        return dict(
            ssm_params=ssm_params,
            dependencies=LambdaInvocationDependenciesWrapperTask(
                lambda_invocation_name=self.lambda_invocation_name,
                manifest_file_path=self.manifest_file_path,
                puppet_account_id=self.puppet_account_id,
                should_use_sns=self.should_use_sns,
                should_use_product_plans=self.should_use_product_plans,
                include_expanded_from=self.include_expanded_from,
                single_account=self.single_account,
                is_dry_run=self.is_dry_run,
                cache_invalidator=self.cache_invalidator,
            ),
        )

    def get_all_params(self):
        all_params = dict()
        for param_name, param_details in self.parameters.items():
            if param_details.get("ssm"):
                with self.input().get("ssm_params").get(param_name).open() as f:
                    all_params[param_name] = json.loads(f.read()).get("Value")
            if param_details.get("default"):
                all_params[param_name] = param_details.get("default")
        return all_params

    def run(self):
        home_region = config.get_home_region(self.puppet_account_id)
        with betterboto_client.CrossAccountClientContextManager(
            "lambda",
            f"arn:aws:iam::{self.puppet_account_id}:role/servicecatalog-puppet/PuppetRole",
            f"sc-{home_region}-{self.puppet_account_id}",
            region_name=home_region,
        ) as lambda_client:
            payload = dict(
                account_id=self.account_id,
                region=self.region,
                parameters=self.get_all_params(),
            )
            response = lambda_client.invoke(
                FunctionName=self.function_name,
                InvocationType=self.invocation_type,
                Payload=json.dumps(payload),
                Qualifier=self.qualifier,
            )
        success_results = dict(RequestResponse=200, Event=202, DryRun=204)

        if success_results.get(self.invocation_type) != response.get("StatusCode"):
            raise Exception(
                f"{self.lambda_invocation_name} failed for {self.account_id}, {self.region}"
            )
        else:
            if response.get("FunctionError"):
                error_payload = response.get("Payload").read()
                raise Exception(error_payload)
            else:
                output = dict(
                    **self.params_for_results_display(),
                    payload=payload,
                    response=response,
                )
                self.write_output(output)


class LambdaInvocationDependenciesWrapperTask(
    workflow_tasks.PuppetTask, manifest_tasks.ManifestMixen
):
    lambda_invocation_name = luigi.Parameter()
    manifest_file_path = luigi.Parameter()

    puppet_account_id = luigi.Parameter()
    should_use_sns = luigi.BoolParameter()
    should_use_product_plans = luigi.BoolParameter()
    include_expanded_from = luigi.BoolParameter()
    single_account = luigi.Parameter()
    is_dry_run = luigi.BoolParameter()
    cache_invalidator = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "manifest_file_path": self.manifest_file_path,
            "lambda_invocation_name": self.lambda_invocation_name,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        lambda_invocation = self.manifest.get("lambda-invocations").get(
            self.lambda_invocation_name
        )

        dependencies = list()
        for dependency in lambda_invocation.get("depends_on", []):
            if isinstance(dependency, str):
                dependencies.append(
                    LaunchTask(
                        launch_name=dependency,
                        manifest_file_path=self.manifest_file_path,
                        puppet_account_id=self.puppet_account_id,
                        should_use_sns=self.should_use_sns,
                        should_use_product_plans=self.should_use_product_plans,
                        include_expanded_from=self.include_expanded_from,
                        single_account=self.single_account,
                        is_dry_run=self.is_dry_run,
                        execution_mode="hub",
                        cache_invalidator=self.cache_invalidator,
                    )
                )
            else:
                dependency_type = dependency.get("type", "launch")
                if dependency_type == "launch":
                    dependencies.append(
                        LaunchTask(
                            launch_name=dependency.get("name"),
                            manifest_file_path=self.manifest_file_path,
                            puppet_account_id=self.puppet_account_id,
                            should_use_sns=self.should_use_sns,
                            should_use_product_plans=self.should_use_product_plans,
                            include_expanded_from=self.include_expanded_from,
                            single_account=self.single_account,
                            is_dry_run=self.is_dry_run,
                            execution_mode="hub",
                            cache_invalidator=self.cache_invalidator,
                        )
                    )
                elif dependency_type == "lambda-invocation":
                    dependencies.append(
                        LambdaInvocationTask(
                            lambda_invocation_name=dependency.get("name"),
                            manifest_file_path=self.manifest_file_path,
                            puppet_account_id=self.puppet_account_id,
                            should_use_sns=self.should_use_sns,
                            should_use_product_plans=self.should_use_product_plans,
                            include_expanded_from=self.include_expanded_from,
                            single_account=self.single_account,
                            is_dry_run=self.is_dry_run,
                            cache_invalidator=self.cache_invalidator,
                        )
                    )
        return dependencies

    def run(self):
        self.write_output(self.params_for_results_display())


class LambdaInvocationTask(workflow_tasks.PuppetTask, manifest_tasks.ManifestMixen):
    lambda_invocation_name = luigi.Parameter()
    manifest_file_path = luigi.Parameter()

    puppet_account_id = luigi.Parameter()
    should_use_sns = luigi.BoolParameter()
    should_use_product_plans = luigi.BoolParameter()
    include_expanded_from = luigi.BoolParameter()
    single_account = luigi.Parameter()
    is_dry_run = luigi.BoolParameter()

    cache_invalidator = luigi.Parameter()

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "manifest_file_path": self.manifest_file_path,
            "lambda_invocation_name": self.lambda_invocation_name,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        requirements = list()
        lambda_invocation = self.manifest.get("lambda-invocations").get(
            self.lambda_invocation_name
        )

        task_defs = self.manifest.get_task_defs_from_details(
            self.puppet_account_id,
            True,
            self.lambda_invocation_name,
            {},
            "lambda-invocations",
        )

        common_params = {
            "lambda_invocation_name": self.lambda_invocation_name,
            "function_name": lambda_invocation.get("function_name"),
            "qualifier": lambda_invocation.get("qualifier", "$LATEST"),
            "invocation_type": lambda_invocation.get("invocation_type"),
            "puppet_account_id": self.puppet_account_id,
            "parameters": lambda_invocation.get("parameters", {}),
            "launch_parameters": lambda_invocation.get("parameters", {}),
            "manifest_parameters": self.manifest.get("parameters", {}),
        }

        wrapper_params = dict(
            manifest_file_path=self.manifest_file_path,
            should_use_sns=self.should_use_sns,
            should_use_product_plans=self.should_use_product_plans,
            include_expanded_from=self.include_expanded_from,
            single_account=self.single_account,
            is_dry_run=self.is_dry_run,
        )

        for task_def in task_defs:
            task_def_parameters = {
                "account_id": task_def.get("account_id"),
                "region": task_def.get("region"),
                "account_parameters": task_def.get("account_parameters"),
                "cache_invalidator": self.cache_invalidator,
            }
            task_def_parameters.update(common_params)
            task_def_parameters.update(wrapper_params)
            requirements.append(InvokeLambdaTask(**task_def_parameters))
        return requirements

    def run(self):
        self.write_output(self.params_for_results_display())


class LambdaInvocationsSectionTask(manifest_tasks.SectionTask):
    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "manifest_file_path": self.manifest_file_path,
            "cache_invalidator": self.cache_invalidator,
        }

    def requires(self):
        requirements = dict(
            invocations=[
                LambdaInvocationTask(
                    lambda_invocation_name=lambda_invocation_name,
                    manifest_file_path=self.manifest_file_path,
                    puppet_account_id=self.puppet_account_id,
                    should_use_sns=self.should_use_sns,
                    should_use_product_plans=self.should_use_product_plans,
                    include_expanded_from=self.include_expanded_from,
                    single_account=self.single_account,
                    is_dry_run=self.is_dry_run,
                    cache_invalidator=self.cache_invalidator,
                )
                for lambda_invocation_name, lambda_invocation in self.manifest.get(
                    "lambda-invocations", {}
                ).items()
            ]
        )
        return requirements

    def run(self):
        self.write_output(self.manifest.get("lambda-invocations"))



from . import tasks_unit_tests


class DeleteCloudFormationStackTaskTest(tasks_unit_tests.PuppetTaskUnitTest):
    account_id = "0123456789010"
    region = "eu-west-0"
    stack_name = "foo"

    def setUp(self) -> None:
        from servicecatalog_puppet.workflow import general

        self.module = general
        self.sut = self.module.DeleteCloudFormationStackTask(
            account_id=self.account_id, region=self.region, stack_name=self.stack_name,
        )

    def test_params_for_results_display(self):
        expected_result = {
            "account_id": self.account_id,
            "region": self.region,
            "stack_name": self.stack_name,
        }
        self.assertEqual(expected_result, self.sut.params_for_results_display())

    def test_api_calls_used(self):
        expected_result = {
            f"cloudformation.describe_stacks_single_page_{self.account_id}_{self.region}": 1,
            f"cloudformation.delete_stack_{self.account_id}_{self.region}": 1,
            f"cloudformation.describe_stack_events_{self.account_id}_{self.region}": 1,
        }
        self.assertEqual(expected_result, self.sut.api_calls_used())



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import click
import yaml

from servicecatalog_puppet import core, config


@click.group()
@click.option("--info/--no-info", default=False)
@click.option("--info-line-numbers/--no-info-line-numbers", default=False)
def cli(info, info_line_numbers):
    """cli for pipeline tools"""
    core.cli(info, info_line_numbers)


@cli.command()
@click.argument("f", type=click.File())
@click.option("--single-account", default=None)
@click.option("--num-workers", default=10)
@click.option("--execution-mode", default="hub")
@click.option("--puppet-account-id", default=None)
@click.option("--home-region", default=None)
@click.option("--regions", default="")
@click.option("--should-collect-cloudformation-events", default=None, type=bool)
@click.option("--should-forward-events-to-eventbridge", default=None, type=bool)
@click.option("--should-forward-failures-to-opscenter", default=None, type=bool)
@click.option("--on-complete-url", default=None)
def deploy(
    f,
    single_account,
    num_workers,
    execution_mode,
    puppet_account_id,
    home_region,
    regions,
    should_collect_cloudformation_events,
    should_forward_events_to_eventbridge,
    should_forward_failures_to_opscenter,
    on_complete_url,
):
    if puppet_account_id is None:
        puppet_account_id = config.get_puppet_account_id()
    executor_account_id = config.get_current_account_id()

    if executor_account_id != puppet_account_id:
        click.echo("Not running in puppet account: writing config")
        with open("config.yaml", "w") as conf:
            conf.write(
                yaml.safe_dump(
                    dict(
                        home_region=home_region,
                        regions=regions.split(","),
                        should_collect_cloudformation_events=should_collect_cloudformation_events,
                        should_forward_events_to_eventbridge=should_forward_events_to_eventbridge,
                        should_forward_failures_to_opscenter=should_forward_failures_to_opscenter,
                    )
                )
            )

    core.deploy(
        f,
        puppet_account_id,
        executor_account_id,
        single_account=single_account,
        num_workers=num_workers,
        execution_mode=execution_mode,
        on_complete_url=on_complete_url,
    )


@cli.command()
@click.argument("f", type=click.File())
@click.option("--single-account", default=None)
def graph(f, single_account):
    core.graph(f)


@cli.command()
@click.argument("f", type=click.File())
@click.option("--single-account", default=None)
@click.option("--puppet-account-id", default=None)
def dry_run(f, single_account, puppet_account_id):
    if puppet_account_id is None:
        puppet_account_id = config.get_puppet_account_id()
    executor_account_id = config.get_current_account_id()
    core.deploy(
        f,
        executor_account_id,
        puppet_account_id,
        single_account=single_account,
        is_dry_run=True,
    )


@cli.command()
@click.argument("puppet_account_id")
@click.argument("iam_role_arns", nargs=-1)
@click.option(
    "--permission-boundary", default="arn:aws:iam::aws:policy/AdministratorAccess"
)
def bootstrap_spoke_as(puppet_account_id, iam_role_arns, permission_boundary):
    core.bootstrap_spoke_as(puppet_account_id, iam_role_arns, permission_boundary)


@cli.command()
@click.argument("puppet_account_id")
@click.option(
    "--permission-boundary", default="arn:aws:iam::aws:policy/AdministratorAccess"
)
def bootstrap_spoke(puppet_account_id, permission_boundary):
    core.bootstrap_spoke(puppet_account_id, permission_boundary)


@cli.command()
@click.argument("ou_path_or_id")
@click.argument("role_name")
@click.argument("iam_role_arns", nargs=-1)
@click.option(
    "--permission-boundary", default="arn:aws:iam::aws:policy/AdministratorAccess"
)
@click.option("--num-workers", default=10)
def bootstrap_spokes_in_ou(
    ou_path_or_id, role_name, iam_role_arns, permission_boundary, num_workers
):
    core.bootstrap_spokes_in_ou(
        ou_path_or_id, role_name, iam_role_arns, permission_boundary, num_workers
    )


@cli.command()
@click.argument("branch-to-bootstrap")
@click.option("--with-manual-approvals/--with-no-manual-approvals", default=False)
@click.option(
    "--puppet-code-pipeline-role-permission-boundary",
    default="arn:aws:iam::aws:policy/AdministratorAccess",
    show_default=True,
)
@click.option(
    "--source-role-permissions-boundary",
    default="arn:aws:iam::aws:policy/AdministratorAccess",
    show_default=True,
)
@click.option(
    "--puppet-generate-role-permission-boundary",
    default="arn:aws:iam::aws:policy/AdministratorAccess",
    show_default=True,
)
@click.option(
    "--puppet-deploy-role-permission-boundary",
    default="arn:aws:iam::aws:policy/AdministratorAccess",
    show_default=True,
)
@click.option(
    "--puppet-provisioning-role-permissions-boundary",
    default="arn:aws:iam::aws:policy/AdministratorAccess",
    show_default=True,
)
@click.option(
    "--cloud-formation-deploy-role-permissions-boundary",
    default="arn:aws:iam::aws:policy/AdministratorAccess",
    show_default=True,
)
@click.option("--deploy_num_workers", default=10, type=click.INT, show_default=True)
@click.option("--source-provider", default="CodeCommit")
@click.option("--repository_name", default="ServiceCatalogPuppet")
@click.option("--branch-name", default="master")
@click.option("--owner")
@click.option("--repo")
@click.option("--branch")
@click.option("--poll-for-source-changes")
@click.option("--webhook-secret")
def bootstrap_branch(
    branch_to_bootstrap,
    with_manual_approvals,
    puppet_code_pipeline_role_permission_boundary,
    source_role_permissions_boundary,
    puppet_generate_role_permission_boundary,
    puppet_deploy_role_permission_boundary,
    puppet_provisioning_role_permissions_boundary,
    cloud_formation_deploy_role_permissions_boundary,
    deploy_num_workers,
    source_provider,
    repository_name,
    branch_name,
    owner,
    repo,
    branch,
    poll_for_source_changes,
    webhook_secret,
):
    puppet_account_id = config.get_puppet_account_id()

    if source_provider == "CodeCommit":
        core.bootstrap_branch(
            branch_to_bootstrap,
            puppet_account_id,
            with_manual_approvals,
            puppet_code_pipeline_role_permission_boundary,
            source_role_permissions_boundary,
            puppet_generate_role_permission_boundary,
            puppet_deploy_role_permission_boundary,
            puppet_provisioning_role_permissions_boundary,
            cloud_formation_deploy_role_permissions_boundary,
            deploy_num_workers,
            source_provider,
            None,
            repository_name,
            branch_name,
            poll_for_source_changes,
            webhook_secret,
        )
    elif source_provider == "GitHub":
        core.bootstrap_branch(
            branch_to_bootstrap,
            puppet_account_id,
            with_manual_approvals,
            puppet_code_pipeline_role_permission_boundary,
            source_role_permissions_boundary,
            puppet_generate_role_permission_boundary,
            puppet_deploy_role_permission_boundary,
            puppet_provisioning_role_permissions_boundary,
            cloud_formation_deploy_role_permissions_boundary,
            deploy_num_workers,
            source_provider,
            owner,
            repo,
            branch,
            poll_for_source_changes,
            webhook_secret,
        )
    else:
        raise Exception(f"Unsupported source provider: {source_provider}")


@cli.command()
@click.option("--with-manual-approvals/--with-no-manual-approvals", default=False)
@click.option(
    "--puppet-code-pipeline-role-permission-boundary",
    default="arn:aws:iam::aws:policy/AdministratorAccess",
    show_default=True,
)
@click.option(
    "--source-role-permissions-boundary",
    default="arn:aws:iam::aws:policy/AdministratorAccess",
    show_default=True,
)
@click.option(
    "--puppet-generate-role-permission-boundary",
    default="arn:aws:iam::aws:policy/AdministratorAccess",
    show_default=True,
)
@click.option(
    "--puppet-deploy-role-permission-boundary",
    default="arn:aws:iam::aws:policy/AdministratorAccess",
    show_default=True,
)
@click.option(
    "--puppet-provisioning-role-permissions-boundary",
    default="arn:aws:iam::aws:policy/AdministratorAccess",
    show_default=True,
)
@click.option(
    "--cloud-formation-deploy-role-permissions-boundary",
    default="arn:aws:iam::aws:policy/AdministratorAccess",
    show_default=True,
)
@click.option(
    "--deploy_environment_compute_type",
    default="BUILD_GENERAL1_SMALL",
    show_default=True,
)
@click.option("--deploy_num_workers", default=10, type=click.INT, show_default=True)
@click.option("--source-provider", default="CodeCommit")
@click.option("--repository_name", default="ServiceCatalogPuppet")
@click.option("--branch-name", default="master")
@click.option("--owner")
@click.option("--repo")
@click.option("--branch")
@click.option("--poll-for-source-changes")
@click.option("--webhook-secret")
def bootstrap(
    with_manual_approvals,
    puppet_code_pipeline_role_permission_boundary,
    source_role_permissions_boundary,
    puppet_generate_role_permission_boundary,
    puppet_deploy_role_permission_boundary,
    puppet_provisioning_role_permissions_boundary,
    cloud_formation_deploy_role_permissions_boundary,
    deploy_environment_compute_type,
    deploy_num_workers,
    source_provider,
    repository_name,
    branch_name,
    owner,
    repo,
    branch,
    poll_for_source_changes,
    webhook_secret,
):
    puppet_account_id = config.get_puppet_account_id()
    if source_provider == "CodeCommit":
        core.bootstrap(
            with_manual_approvals,
            puppet_account_id,
            puppet_code_pipeline_role_permission_boundary,
            source_role_permissions_boundary,
            puppet_generate_role_permission_boundary,
            puppet_deploy_role_permission_boundary,
            puppet_provisioning_role_permissions_boundary,
            cloud_formation_deploy_role_permissions_boundary,
            deploy_environment_compute_type,
            deploy_num_workers,
            source_provider,
            None,
            repository_name,
            branch_name,
            poll_for_source_changes,
            webhook_secret,
        )
    elif source_provider == "GitHub":
        core.bootstrap(
            with_manual_approvals,
            puppet_account_id,
            puppet_code_pipeline_role_permission_boundary,
            source_role_permissions_boundary,
            puppet_generate_role_permission_boundary,
            puppet_deploy_role_permission_boundary,
            puppet_provisioning_role_permissions_boundary,
            cloud_formation_deploy_role_permissions_boundary,
            deploy_environment_compute_type,
            deploy_num_workers,
            source_provider,
            owner,
            repo,
            branch,
            poll_for_source_changes,
            webhook_secret,
        )
    else:
        raise Exception(f"Unsupported source provider: {source_provider}")


@cli.command()
@click.argument("complexity", default="simple")
@click.argument("p", type=click.Path(exists=True))
def seed(complexity, p):
    core.seed(complexity, p)


@cli.command()
@click.argument("expanded_manifest", type=click.File())
@click.option("--format", "-f", type=click.Choice(["table", "json"]), default="table")
def list_launches(expanded_manifest, format):
    current_account_id = puppet_account_id = config.get_puppet_account_id()
    core.deploy(
        expanded_manifest,
        puppet_account_id,
        current_account_id,
        single_account=None,
        is_dry_run=True,
        is_list_launches=format,
    )


@cli.command()
@click.argument("f", type=click.File())
def expand(f):
    core.expand(f)


@cli.command()
@click.argument("f", type=click.File())
def validate(f):
    core.validate(f)


@cli.command()
def version():
    core.version()


@cli.command()
@click.argument("p", type=click.Path(exists=True))
def upload_config(p):
    content = open(p, "r").read()
    config = yaml.safe_load(content)
    core.upload_config(config)


@cli.command()
@click.argument("org-iam-role-arn")
def set_org_iam_role_arn(org_iam_role_arn):
    core.set_org_iam_role_arn(org_iam_role_arn)


@cli.command()
@click.argument("puppet_account_id")
def bootstrap_org_master(puppet_account_id):
    core.bootstrap_org_master(puppet_account_id)


@cli.command()
@click.argument("what", default="puppet")
@click.option("--tail/--no-tail", default=False)
def run(what, tail):
    core.run(what, tail)


@cli.command()
def list_resources():
    core.list_resources()


@cli.command()
@click.argument("f", type=click.File())
@click.argument("name")
@click.argument("portfolio_name")
def import_product_set(f, name, portfolio_name):
    core.import_product_set(f, name, portfolio_name)


@cli.command()
@click.argument("account_or_ou_file_path", type=click.File())
def add_to_accounts(account_or_ou_file_path):
    core.add_to_accounts(yaml.safe_load(account_or_ou_file_path))


@cli.command()
@click.argument("account_id_or_ou_id_or_ou_path")
def remove_from_accounts(account_id_or_ou_id_or_ou_path):
    core.remove_from_accounts(account_id_or_ou_id_or_ou_path)


@cli.command()
@click.argument("launch_file_path", type=click.File())
def add_to_launches(launch_name, launch_file_path):
    core.add_to_launches(launch_name, yaml.safe_load(launch_file_path))


@cli.command()
@click.argument("launch_name")
def remove_from_launches(launch_name):
    core.remove_from_launches(launch_name)


@cli.command()
@click.argument("f", type=click.File())
def reset_provisioned_product_owner(f):
    core.reset_provisioned_product_owner(f)


@cli.command()
@click.argument("regions", nargs=-1)
def set_regions(regions):
    core.set_config_value("regions", regions)


@cli.command()
@click.argument("name")
@click.argument("value")
def set_config_value(name, value):
    core.set_config_value(name, value)


@cli.command()
@click.argument("execution_id")
@click.option("--puppet-account-id", default=None)
def export_puppet_pipeline_logs(execution_id, puppet_account_id):
    if puppet_account_id is None:
        puppet_account_id = config.get_puppet_account_id()
    core.export_puppet_pipeline_logs(execution_id, puppet_account_id)


@cli.command()
@click.option("--puppet-account-id", default=None)
def uninstall(puppet_account_id):
    if puppet_account_id is None:
        puppet_account_id = config.get_puppet_account_id()
    core.uninstall(puppet_account_id)


@cli.command()
@click.option("--puppet-account-id", default=None)
def release_spoke(puppet_account_id):
    if puppet_account_id is None:
        puppet_account_id = config.get_puppet_account_id()
    core.release_spoke(puppet_account_id)


@cli.command()
@click.argument("iam_role_arns", nargs=-1)
def wait_for_code_build_in(iam_role_arns):
    core.wait_for_code_build_in(iam_role_arns)
    click.echo("AWS CodeBuild is available")


if __name__ == "__main__":
    cli()



import configparser
import os

import yaml
import logging
import json
from copy import deepcopy

from servicecatalog_puppet import config
from servicecatalog_puppet.macros import macros
from servicecatalog_puppet import constants

logger = logging.getLogger(__file__)


def load(f, puppet_account_id):
    manifest_name = f.name
    manifest = {
        "schema": "puppet-2019-04-01",
        "parameters": {},
        "accounts": [],
        "launches": {},
        "spoke-local-portfolios": {},
    }
    manifest.update(yaml.safe_load(f.read()))
    d = os.path.dirname(os.path.abspath(f.name))

    extendable = ["parameters", "launches", "spoke-local-portfolios"]
    for t in extendable:
        t_path = f"{d}{os.path.sep}{t}"
        if os.path.exists(t_path):
            for f in os.listdir(t_path):
                with open(f"{t_path}{os.path.sep}{f}", "r") as file:
                    manifest[t].update(yaml.safe_load(file.read()))

    if os.path.exists(f"{d}{os.path.sep}manifests"):
        for f in os.listdir(f"{d}{os.path.sep}manifests"):
            with open(f"{d}{os.path.sep}manifests{os.path.sep}{f}", "r") as file:
                ext = yaml.safe_load(file.read())
                for t in extendable:
                    manifest[t].update(ext.get(t, {}))

    for config_file in [
        manifest_name.replace(".yaml", ".properties"),
        manifest_name.replace(".yaml", f"-{puppet_account_id}.properties"),
    ]:
        parser = configparser.ConfigParser(
            interpolation=configparser.BasicInterpolation()
        )
        parser.optionxform = str

        if os.path.exists(config_file):
            logger.info(f"reading {config_file}")
            parser.read(config_file)
            for name, value in parser["launches"].items():
                launch_name, property_name = name.split(".")
                if property_name != "version":
                    raise Exception(
                        "You can only specify a version in the properties file"
                    )
                manifest["launches"][launch_name][property_name] = value
    for launch_name, launch_details in manifest.get("launches").items():
        if launch_details.get("execution") is None:
            launch_details["execution"] = constants.EXECUTION_MODE_DEFAULT
    return manifest


def expand_manifest(manifest, client):
    new_manifest = deepcopy(manifest)
    temp_accounts = []

    logger.info("Starting the expand")

    for account in manifest.get("accounts"):
        if account.get("account_id"):
            logger.info("Found an account: {}".format(account.get("account_id")))
            temp_accounts.append(account)
        elif account.get("ou"):
            ou = account.get("ou")
            logger.info("Found an ou: {}".format(ou))
            if ou.startswith("/"):
                temp_accounts += expand_path(account, client)
            else:
                temp_accounts += expand_ou(account, client)

    for parameter_name, parameter_details in new_manifest.get("parameters", {}).items():
        if parameter_details.get("macro"):
            macro_to_run = macros.get(parameter_details.get("macro").get("method"))
            result = macro_to_run(client, parameter_details.get("macro").get("args"))
            parameter_details["default"] = result
            del parameter_details["macro"]

    accounts_by_id = {}
    for account in temp_accounts:
        for parameter_name, parameter_details in account.get("parameters", {}).items():
            if parameter_details.get("macro"):
                macro_to_run = macros.get(parameter_details.get("macro").get("method"))
                result = macro_to_run(
                    client, parameter_details.get("macro").get("args")
                )
                parameter_details["default"] = result
                del parameter_details["macro"]

        account_id = account.get("account_id")
        if account.get("append") or account.get("overwrite"):
            if (
                account.get("default_region")
                or account.get("regions_enabled")
                or account.get("tags")
            ):
                raise Exception(
                    f"{account_id}: If using append or overwrite you cannot set default_region, regions_enabled or tags"
                )

        if accounts_by_id.get(account_id) is None:
            accounts_by_id[account_id] = account
        else:
            stored_account = accounts_by_id[account_id]
            stored_account.update(account)

            if stored_account.get("append"):
                append = stored_account.get("append")
                for tag in append.get("tags", []):
                    stored_account.get("tags").append(tag)
                for region_enabled in append.get("regions_enabled", []):
                    stored_account.get("regions_enabled").append(region_enabled)
                del stored_account["append"]

            elif stored_account.get("overwrite"):
                overwrite = stored_account.get("overwrite")
                if overwrite.get("tags"):
                    stored_account["tags"] = overwrite.get("tags")
                if overwrite.get("regions_enabled"):
                    stored_account["regions_enabled"] = overwrite.get("regions_enabled")
                if overwrite.get("default_region"):
                    stored_account["default_region"] = overwrite.get("default_region")
                del stored_account["overwrite"]

            else:
                raise Exception(
                    f"Account {account_id} has been seen twice without using append or overwrite"
                )
    new_manifest["accounts"] = list(accounts_by_id.values())

    for launch_name, launch_details in new_manifest.get(constants.LAUNCHES, {}).items():
        for parameter_name, parameter_details in launch_details.get(
            "parameters", {}
        ).items():
            if parameter_details.get("macro"):
                macro_to_run = macros.get(parameter_details.get("macro").get("method"))
                result = macro_to_run(
                    client, parameter_details.get("macro").get("args")
                )
                parameter_details["default"] = result
                del parameter_details["macro"]

    return new_manifest


def expand_path(account, client):
    ou = client.convert_path_to_ou(account.get("ou"))
    account["ou"] = ou
    return expand_ou(account, client)


def expand_ou(original_account, client):
    expanded = []
    exclusions = original_account.get("exclude", {}).get("accounts", [])
    ou_exclusions = original_account.get("exclude", {}).get("ous", [])
    for ou_exclusion in ou_exclusions:
        if ou_exclusion.startswith("/"):
            ou_id = client.convert_path_to_ou(ou_exclusion)
        else:
            ou_id = ou_exclusion
        children = client.list_children_nested(ParentId=ou_id, ChildType="ACCOUNT")
        for child in children:
            logger.info(
                f"Adding {child.get('Id')} to the exclusion list as it was in the ou {ou_exclusion}"
            )
            exclusions.append(child.get("Id"))

    response = client.list_children_nested(
        ParentId=original_account.get("ou"), ChildType="ACCOUNT"
    )
    for result in response:
        new_account_id = result.get("Id")
        if new_account_id in exclusions:
            logger.info(f"Skipping {new_account_id} as it is in the exclusion list")
            continue
        response = client.describe_account(AccountId=new_account_id)
        new_account = deepcopy(original_account)
        del new_account["ou"]
        if response.get("Account").get("Status") == "ACTIVE":
            if response.get("Account").get("Name") is not None:
                new_account["name"] = response.get("Account").get("Name")
            new_account["email"] = response.get("Account").get("Email")
            new_account["account_id"] = new_account_id
            new_account["expanded_from"] = original_account.get("ou")
            new_account["organization"] = (
                response.get("Account").get("Arn").split(":")[5].split("/")[1]
            )
            expanded.append(new_account)
        else:
            logger.info(
                f"Skipping account as it is not ACTIVE: {json.dumps(response.get('Account'), default=str)}"
            )
    return expanded


def get_configuration_overrides(manifest, launch_details):
    configuration = dict()
    if manifest.get("configuration"):
        if manifest.get("configuration").get("retry_count"):
            configuration["retry_count"] = manifest.get("configuration").get(
                "retry_count"
            )
    if launch_details.get("configuration"):
        if launch_details.get("configuration").get("retry_count"):
            configuration["retry_count"] = launch_details.get("configuration").get(
                "retry_count"
            )
        if launch_details.get("configuration").get("requested_priority"):
            configuration["requested_priority"] = int(
                launch_details.get("configuration").get("requested_priority")
            )
    return configuration


class Manifest(dict):
    def get_actions_from(
        self, launch_name, pre_or_post, launch_or_spoke_local_portfolio
    ):
        logger.info(
            f"get_actions_from for {launch_or_spoke_local_portfolio}.{launch_name}"
        )
        launch_details = self.get(launch_or_spoke_local_portfolio).get(launch_name)
        actions = self.get("actions")
        result = list()
        for provision_action in launch_details.get(f"{pre_or_post}_actions", []):
            action = dict()
            action.update(actions.get(provision_action.get("name")))
            action.update(provision_action)
            action["source"] = launch_name
            action["phase"] = pre_or_post
            action["source_type"] = launch_or_spoke_local_portfolio
            result.append(action)
        return result

    def get_account(self, account_id):
        for account in self.get("accounts"):
            if account.get("account_id") == account_id:
                return account
        raise Exception(f"Could not find account: {account_id}")

    def get_sharing_policies_by_region(self):
        sharing_policies_by_region = {}
        for account in self.get("accounts"):
            account_regions = list()
            if account.get("default_region") is None:
                raise Exception(
                    f"Account {account.get('account_id')} has no default_region"
                )
            account_regions.append(account.get("default_region"))

            enabled_regions = (
                account.get("enabled", [])
                + account.get("regions_enabled", [])
                + account.get("enabled_regions", [])
            )
            if len(enabled_regions) == 0:
                raise Exception(
                    f"Account {account.get('account_id')} has no enabled|regions_enabled|enabled_regions"
                )
            account_regions += enabled_regions

            for r in account_regions:
                if sharing_policies_by_region.get(r) is None:
                    sharing_policies_by_region[r] = dict(accounts=[], organizations=[])
                if account.get("organization", "") != "":
                    organization = account.get("organization")
                    if (
                        organization
                        not in sharing_policies_by_region[r]["organizations"]
                    ):
                        sharing_policies_by_region[r]["organizations"].append(
                            organization
                        )
                else:
                    account_id = account.get("account_id")
                    if account_id not in sharing_policies_by_region[r]["accounts"]:
                        sharing_policies_by_region[r]["accounts"].append(account_id)

        return sharing_policies_by_region

    def get_shares_by_region_portfolio_account(self, puppet_account_id, section):
        shares_by_region_portfolio_account = {}
        configuration = {}
        include_expanded_from = False
        for launch_name, launch_details in self.get(section, {}).items():
            portfolio = launch_details.get("portfolio")
            tasks = self.get_task_defs_from_details(
                puppet_account_id,
                include_expanded_from,
                launch_name,
                configuration,
                section,
            )
            for task in tasks:
                account_id = task.get("account_id")
                region = task.get("region")
                if shares_by_region_portfolio_account.get(region) is None:
                    shares_by_region_portfolio_account[region] = {}
                if shares_by_region_portfolio_account[region].get(portfolio) is None:
                    shares_by_region_portfolio_account[region][portfolio] = {}
                result = self.get_account(account_id)
                result[section] = launch_details
                shares_by_region_portfolio_account[region][portfolio][
                    account_id
                ] = result
        return shares_by_region_portfolio_account

    def get_accounts_by_region(self):
        accounts_by_region = {}
        for account in self.get("accounts"):
            account_regions = list()
            account_regions += account.get("enabled", [])
            account_regions += account.get("regions_enabled", [])
            account_regions += account.get("enabled_regions", [])
            account_regions.append(account.get("default_region"))
            for r in account_regions:
                if accounts_by_region.get(r) is None:
                    accounts_by_region[r] = []
                accounts_by_region[r].append(account)
        return accounts_by_region

    def get_task_defs_from_details(
        self,
        puppet_account_id,
        include_expanded_from,
        launch_name,
        configuration,
        launch_or_spoke_local_portfolio,
    ):
        launch_details = self.get(launch_or_spoke_local_portfolio).get(launch_name)
        accounts = self.get("accounts")
        if launch_details is None:
            raise Exception(f"launch_details is None for {launch_name}")
        if launch_or_spoke_local_portfolio == "lambda-invocations":
            deploy_to = launch_details.get("invoke_for")
        elif launch_or_spoke_local_portfolio == "launches":
            deploy_to = launch_details.get("deploy_to")
        elif launch_or_spoke_local_portfolio == "spoke-local-portfolios":
            deploy_to = launch_details.get("deploy_to") or launch_details.get(
                "share_with"
            )
        task_defs = []
        for tag_list_item in deploy_to.get("tags", []):
            for account in accounts:
                for tag in account.get("tags", []):
                    if tag == tag_list_item.get("tag"):
                        tag_account_def = deepcopy(configuration)
                        tag_account_def["account_id"] = account.get("account_id")
                        if include_expanded_from:
                            tag_account_def["expanded_from"] = account.get(
                                "expanded_from"
                            )
                            tag_account_def["organization"] = account.get(
                                "organization"
                            )
                        tag_account_def["account_parameters"] = account.get(
                            "parameters", {}
                        )

                        regions = tag_list_item.get("regions", "default_region")
                        if isinstance(regions, str):
                            if regions in [
                                "enabled",
                                "regions_enabled",
                                "enabled_regions",
                            ]:
                                for region_enabled in account.get("regions_enabled"):
                                    region_tag_account_def = deepcopy(tag_account_def)
                                    region_tag_account_def["region"] = region_enabled
                                    task_defs.append(region_tag_account_def)
                            elif regions == "default_region":
                                region_tag_account_def = deepcopy(tag_account_def)
                                region_tag_account_def["region"] = account.get(
                                    "default_region"
                                )
                                task_defs.append(region_tag_account_def)
                            elif regions == "all":
                                all_regions = config.get_regions(puppet_account_id)
                                for region_enabled in all_regions:
                                    region_tag_account_def = deepcopy(tag_account_def)
                                    region_tag_account_def["region"] = region_enabled
                                    task_defs.append(region_tag_account_def)
                            else:
                                raise Exception(
                                    f"Unsupported regions {regions} setting for launch: {launch_name}"
                                )
                        elif isinstance(regions, list):
                            for region in regions:
                                region_tag_account_def = deepcopy(tag_account_def)
                                region_tag_account_def["region"] = region
                                task_defs.append(region_tag_account_def)
                        elif isinstance(regions, tuple):
                            for region in regions:
                                region_tag_account_def = deepcopy(tag_account_def)
                                region_tag_account_def["region"] = region
                                task_defs.append(region_tag_account_def)
                        else:
                            raise Exception(
                                f"Unexpected regions of {regions} set for launch {launch_name}"
                            )

        for account_list_item in deploy_to.get("accounts", []):
            for account in accounts:
                if account.get("account_id") == account_list_item.get("account_id"):
                    account_account_def = deepcopy(configuration)
                    account_account_def["account_id"] = account.get("account_id")
                    if include_expanded_from:
                        account_account_def["expanded_from"] = account.get(
                            "expanded_from"
                        )
                        account_account_def["organization"] = account.get(
                            "organization"
                        )
                    account_account_def["account_parameters"] = account.get(
                        "parameters", {}
                    )

                    regions = account_list_item.get("regions", "default_region")
                    if isinstance(regions, str):
                        if regions in ["enabled", "regions_enabled", "enabled_regions"]:
                            for region_enabled in account.get("regions_enabled"):
                                region_account_account_def = deepcopy(
                                    account_account_def
                                )
                                region_account_account_def["region"] = region_enabled
                                task_defs.append(region_account_account_def)
                        elif regions in ["default_region"]:
                            region_account_account_def = deepcopy(account_account_def)
                            region_account_account_def["region"] = account.get(
                                "default_region"
                            )
                            task_defs.append(region_account_account_def)
                        elif regions in ["all"]:
                            all_regions = config.get_regions(puppet_account_id)
                            for region_enabled in all_regions:
                                region_account_account_def = deepcopy(
                                    account_account_def
                                )
                                region_account_account_def["region"] = region_enabled
                                task_defs.append(region_account_account_def)
                        else:
                            raise Exception(
                                f"Unsupported regions {regions} setting for launch: {launch_name}"
                            )

                    elif isinstance(regions, list):
                        for region in regions:
                            region_account_account_def = deepcopy(account_account_def)
                            region_account_account_def["region"] = region
                            task_defs.append(region_account_account_def)
                    elif isinstance(regions, tuple):
                        for region in regions:
                            region_account_account_def = deepcopy(account_account_def)
                            region_account_account_def["region"] = region
                            task_defs.append(region_account_account_def)
                    else:
                        raise Exception(
                            f"Unexpected regions of {regions} set for launch {launch_name}"
                        )
        return task_defs
